# UpdateManager.py
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-
import os
import apt_pkg
import sys
import time
import shutil
import json
import dbus
import logging
import dbus.service
import subprocess
import traceback
from gettext import gettext as _
from apt.debfile import DebPackage
from dbus.mainloop.glib import DBusGMainLoop
import collections
DBusGMainLoop(set_as_default=True)
from .Core.errors import *
from .Core.enums import *
from .Core.MyCache import MyCache
from .UpdateManagerDbus import UpdateManagerDbusController,UpdateManagerDbusControllerUtils,UPDATER_DBUS_INTERFACE,UPDATER_DBUS_PATH,UPDATER_DBUS_SERVICE
from .Core.UpdateList import UpdateList
from .UpdatesAvailable import UpdatesAvailable
from .Core.PluginManager import *
from .backend import InstallBackend,get_backend
from .backend import InhibitShutdownLock
from .Core.Database import Sqlite3Server
from .Core.loop import mainloop
from .Core.DataAcquisition import UpdateMsgCollector
from .Core.DeltaDownload import DeltaDownloadBackend

from SystemUpdater.Core.UpdaterConfigParser import UpgradeConfig
from SystemUpdater.Core.JsonConfigParser import JsonConfig as json_config
from SystemUpdater.Core.utils import deb_verify,PolicyKit_Authority,get_proc_from_dbus_name,get_dist,ApplicationSecInstall,is_offline_source
from aptdaemon.enums import ERROR_DAEMON_DIED

IMPORTANT_FIX_ENV = "allow-fix-env"
OPENKYLIN_DISTTRIBUTOR = "Openkylin"
BACKEND_PKG_NAME = 'kylin-system-updater'
FRONTEND_PKG_NAME = "kylin-update-frontend"
SYSTEM_UPDATE_GROUPS = "kylin-update-desktop-system"
GROUPS_PKG_NAME = 'kylin-update-desktop-config'
SOURCES_UPDATE_NAME = "kylin-software-properties"
APTD_PKG_NAME = "aptdaemon"
IMPORTANT_LIST_PATH = '/var/lib/kylin-software-properties/template/important.list'

APT_PRIVATE_PATH = "/etc/kylin-system-updater/apt_private.conf"
APT_STANDARD_PATH = "/etc/kylin-system-updater/apt_standard.conf"

PACKAGES_DATABASE_PATH = "/usr/share/kylin-update-desktop-config/version/packages-database.json"
STANDARD_PACKAGES_PATH = "/usr/share/kylin-update-desktop-config/version/standard-system.json"
STANDARD_SOURCES_PATH = "/usr/share/kylin-update-desktop-config/version/standard-sources.json"

class UpdateManager():
    def __init__(self,options):
        try:
            self.options = options
            self.cache = None
            self.update_list = None
            self.resolver_backend = None

            self.init_config_aptdeamon = False
            self.aptd_lang_switch = False
            self.now_working = InstallBackend.ACTION_DEFUALT_STATUS
            self.task_queue = collections.deque()

            self._check_exception_status()
            self.dist = get_dist()
            self.bus = dbus.SystemBus()
            
            #config and 数据库
            self.configs_uncover = UpgradeConfig(defaults_dir="system-updater-defaults.conf")
            self.configs_cover = UpgradeConfig(name = "system-updater-coverable.conf")
            self.uuconfigs = UpgradeConfig(datadir = "/var/lib/unattended-upgrades/", name = "unattended-upgrades-policy.conf")
            self.sqlite3_server = Sqlite3Server(self)
            json_config()

            #dbus
            self.dbusController = self._setup_dbus(self.bus)
            self.dbusControllerUtils = self._setup_dbus_utils(self.bus)

            if not self.options.debug:
                self.dbusController.make_aptdeamon_restart()
                # 等待aptdeamon重启
                time.sleep(0.5)

            #数据采集器
            self.collector = UpdateMsgCollector(self)
            self.apt_p2p_config = AptP2pConfigManager()
            self.delay_shutdown = DelayShutdownLock(self)

            self.last_install_mode = True

            #更新安装数据 
            if os.path.exists(UpdatesAvailable.UPDATE_AVAILABLE_DATA):
                local_data = UpdatesAvailable(self)
                local_data.read_local()
                if local_data.available:
                    self.task_queue.append(local_data)

            self.delta = DeltaDownloadBackend(self)
        except Exception as e:
            logging.error(e)
            tbk = traceback.format_exc()
            logging.error(str(tbk))

    def run(self):
        """Start the daemon and listen for calls."""
        logging.info("Waiting for calls...")
        try:
            mainloop.run()
        except KeyboardInterrupt:
            self.dbusController.Quit(None)

    def _check_exception_status(self):
        logging.info("Check: Abonormal Device Startup conditions...")
        InhibitShutdownLock._check_unlock_sleep()

    def check_exception_fix(self):
        try:
            reboot_num = self.configs_uncover.getWithDefault("SystemStatus", "check_abnormal_flag", 0)
            if reboot_num != 0:
                time.sleep(2)
                self.configs_uncover.setValue("SystemStatus","check_abnormal_flag",str(0))
                self.start_available()
                if self.task_queue:
                    install_data = self.task_queue.pop()
                    install_data.available = False
                    error_code = ERROR_INSTALL_POWER_OUTAGE
                    install_content = install_data.get_content()

                    self._record_finish_status(False,install_content,error_code)
                    self._refresh_uuid(False,error_code)

                    self.dbusController.UpdateInstallFinishedNext(False,True,install_content,error_code)
        except Exception as e:
            logging.error(e)
            tbk = traceback.format_exc()
            logging.error(str(tbk))
        
    @property
    def is_openkylin(self):
        return self.dist == OPENKYLIN_DISTTRIBUTOR

    #进行更新的操作
    def start_update(self,update_mode = InstallBackend.MODE_UPDATE_ALL, fix_env = True):
        try:
            if self.options.check_network and is_offline_source() == False:
                self._check_network()

            if self.options.update_source and self.is_openkylin == False:
                def _update_important_success():
                    _fix_env = fix_env
                    if IMPORTANT_FIX_ENV in self._get_important_data():
                        _fix_env = True
                    self.start_update_cache(fix_env = _fix_env)
                self._update_important(reply_handler=_update_important_success,error_handler=self._update_error_handler)
                return

            self.start_update_cache(update_mode,fix_env)
        except UpdateBaseError as excep:
            self._update_error_handler(excep.code)
        except UpdateProgressExit:
            pass
        except Exception as e:
            logging.error(e)
            tbk = traceback.format_exc()
            logging.error(str(tbk))
            self._update_error_handler(ERROR_PROGRAM_EXCEPTION)

    def start_update_cache(self,update_mode = InstallBackend.MODE_UPDATE_ALL,fix_env = True):
        #更新回调
        def _update_finished(parent,success,error_code):
            if success:
                if update_mode == InstallBackend.MODE_UPDATE_ALL:
                    self.start_available(fix_env)
                else:
                    self.dbusController.UpdateDetectFinishedNext(success,[''],error_code)
            else:
                if error_code == ERROR_DAEMON_DIED:
                    self.dbusController.UpdateDetectFinishedNext(success,[BACKEND_PKG_NAME],ERROR_DAEMON_DIED)
                elif error_code == ERROR_NOT_SOURCES_PATH:
                    self.dbusController.UpdateDetectFinishedNext(True,[],'')
                else:
                    self._update_error_handler(error_code)

        self.check_config_aptdeamon()
        update_backend = get_backend(self, InstallBackend.ACTION_UPDATE,update_mode)
        update_backend.connect("finished", _update_finished)
        update_backend.start()

    def start_available(self,fix_env = True):
        try:
            self.configs_cover.reRead()
            self.configs_uncover.reRead()
            self.uuconfigs.reRead()
            json_config().read()

            self.refresh_cache(True,True)

            self.update_list = UpdateList(self)

            self._check_failed_packages(self.cache)

            self._check_system_incomplete(self.cache,fix_env)
            
            self._check_system_broken(self.cache,fix_env)

            # self._check_system_library()
            
            self._check_config_upgrade(self.cache,fix_env)

            self._check_self_upgrade(self.cache,fix_env)

            # ↑ ↑ ↑ ↑ 以上代码无法自我更新修复 不要随意改动添加

            self._check_other_upgrade(self.cache,fix_env)

            self.update_list.update_kylin(self.cache,self._get_important_data(),self.is_openkylin)

            self._update_last_install(self._get_important_data(),self.update_list.get_push())
            self.update_list.backup_serverid()

            # 自动更新和更新提示 都会来检查此处 进行到此处 执行更新前版本匹配检查
            if self.update_list.get_push():
                success,desc = self._check_push_state(self.update_list)
                if not success:
                    # 控制面板 进行弹窗提示
                    if fix_env:
                        self.dbusController.ForceUpdatePopupRequestNext(True,self.update_list.get_push(),desc)
                    # 其他进程调用 直接发送失败
                    else:
                        self.dbusController.UpdateDetectFinishedNext(False,[""],'')
                else:
                    self.dbusController.UpdateDetectFinishedNext(True,self.update_list.get_push(),'')
            else:
                self.dbusController.UpdateDetectFinishedNext(True,self.update_list.get_push(),'')

            if fix_env:
                self.task_queue.clear()

        except UpdateBaseError as excep:
            self._update_error_handler(excep.code)
        except UpdateProgressExit:
            pass
        except Exception as e:
            logging.error(e)
            tbk = traceback.format_exc()
            logging.error(str(tbk))
            self._update_error_handler(ERROR_PROGRAM_EXCEPTION)

    def start_available_next(self):
        if self.update_list:
            self.dbusController.UpdateDetectFinishedNext(True,self.update_list.get_push(),'')

    #计算依赖关系模块
    def start_resolver_depends(self,problem = False,push_content = []):
        try:
            self.check_config_aptdeamon()
            self.now_working = InstallBackend.ACTION_CHECK_RESOLVER
            self.refresh_cache(True,False)
            
            resolver_backend = UpdatesAvailable(self)
            resolver_backend.start(problem,push_content,self.update_list)

            self.task_queue.append(resolver_backend)
        except UpdateBaseError as excep:
            if not problem:
                self.dbusController.UpdateDependResloveStatus(False,False,[''],[''],[''],excep.header,excep.desc)
            else:
                self.dbusController.DistupgradeDependResloveStatus(False,False,[''],[''],[''],excep.header,excep.desc)
        except UpdateProgressExit as excep:
            pass
        except Exception as e:
            logging.error(e) 
        finally:
            self.now_working = InstallBackend.ACTION_DEFUALT_STATUS
    
    #进行下载的操作
    def start_download(self,upgrade_mode = InstallBackend.MODE_DEFAULT_STATUS,is_auto = False):
        downloading_data = self.task_queue[-1]
        shutdown_install = downloading_data.shutdown_install
        downloading_content = downloading_data.upgrade_singles + downloading_data.upgrade_groups

        #回调处理
        def _download_finished(parent,success,error_code):
            self.task_queue.pop()
            # 更新数据到本地
            if success:
                downloading_data.available = True
                self.task_queue.append(downloading_data)

            if shutdown_install:
                self.dbusController.UpdateDownloadFinishedNext(success,downloading_content,error_code)
                if success:
                    if downloading_data.signal_trigger and self.delay_shutdown.lock():
                        self.delay_shutdown.connect_logind()
            else:
                if is_auto:
                    # 运行时安装模式下 调用下载的接口 不能发送下载完成的信号会导致弹窗触发，所以切换到发送关机安装信号 仅针对自动更新 
                    self.dbusController.UpdateInstallFinished(success,downloading_content,\
                                    get_error_string_from_enum(error_code),get_error_description_from_enum(error_code))
                else:
                    if success:
                        self.start_run_install(downloading_content)
                    else:
                        self.dbusController.UpdateInstallFinishedNext(success,True,downloading_content,error_code)
        try:
            download_backend = get_backend(self,InstallBackend.ACTION_DOWNLOADONLY,upgrade_mode,downloading_data)
            download_backend.connect("finished", _download_finished)
            download_backend.start(downloading_content)
        except Exception as e:
            logging.error(e)

    #运行时安装
    def start_run_install(self,install_content = []):
        if not self.task_queue:
            self.dbusController.UpdateInstallFinishedNext(False,False,[''],ERROR_NOT_UPGRADE_PACKAGES)
            return

        installing_data = self.task_queue[-1]
        install_content = installing_data.get_content()

        def _install_finished(parent,success,error_code):
            self.task_queue.pop()
            self.dbusController.UpdateInstallFinishedNext(success,installing_data.steps_queue == [],install_content,error_code)
            installing_data.available = False
            if success:
                if self.update_list:
                    self.update_list.push_pop(install_content)
                
                # 为空时标识全部的包升级完成
                if self.update_list.get_push():
                    pass
                else:
                    self._refresh_uuid(success,error_code)
                    self._refresh_local_sources()
            else:
                self._refresh_uuid(success,error_code)
        try:
            install_backend = get_backend(self,InstallBackend.ACTION_INSTALL,data = installing_data)
            install_backend.connect("finished", _install_finished)
            install_backend.start(install_content)
        except Exception as e:
            logging.error(e)

    #关机安装
    def start_poweroff(self,state=None,by_dbus=False):
        if not bool(self.check_shutdown_install()):
            self.dbusController.UpdateInstallFinishedNext(False,False,[''],ERROR_NOT_UPGRADE_PACKAGES)
            return

        installing_data = self.task_queue[-1]
        install_content = installing_data.get_content()
        if state != None:
            installing_data.device_status = state

        #安装回调
        def _install_finished(parent,success,error_code):
            if by_dbus and error_code in [ERROR_DEVICE_LOW_BATTERY,ERROR_NOT_DISK_SPACE]:
                self.dbusController.UpdateInstallFinishedNext(success,True,install_content,error_code)
                return

            self.task_queue.pop()
            device_state = "reboot"
            installl_finished = installing_data.steps_queue == []
            if success == False:
                device_state = installing_data.device_status
                installing_data.available = False
            elif success == True and installl_finished:                
                device_state = installing_data.device_status
                installing_data.available = False
                self._refresh_local_sources()
            else:
                installing_data.available = True

            self._record_finish_status(success,install_content,error_code)
            self._refresh_uuid(success,error_code)

            self.dbusController.UpdateInstallFinishedNext(success,installl_finished,install_content,error_code)
            self._system_reboot_shutdown(device_state,self.options.prohibit_shutdown)

        install_backend = get_backend(self, InstallBackend.ACTION_INSTALL_SHUTDOWN,data=installing_data)
        install_backend.connect("finished", _install_finished)
        install_backend.start(install_content)

    #进行修复破损的包的操作 apt install -f 
    def start_fix_broken(self):
        def _fix_finished(parent,success,error_code):
            if success:
                self.start_available()
            else:
                self._update_error_handler(error_code)

        fix_backend = get_backend(self, InstallBackend.ACTION_FIX_BROKEN)
        fix_backend.connect("finished", _fix_finished)
        fix_backend.start()

    #自我升级
    def start_selfupgrade(self,upgrade_content,fix_env=True):
        #自我升级回调 下面不要轻易改动 会引起严重问题
        def _selfupgrade_finished(parent,success,error_code):
            if success:
                channel_config = json_config().getWithDefault("update_channel_upgrade",default = {})
                # restart_list = channel_config.get("restart_service_list",[APTD_PKG_NAME,BACKEND_PKG_NAME])
                notify_list = channel_config.get("restart_panel_list",[FRONTEND_PKG_NAME])
                if APTD_PKG_NAME in upgrade_content and BACKEND_PKG_NAME not in upgrade_content:
                    self.dbusController.make_aptdeamon_restart()
                if FRONTEND_PKG_NAME in upgrade_content and FRONTEND_PKG_NAME in notify_list:
                    self.dbusController.UpdateDetectFinishedNext(False,[BACKEND_PKG_NAME],PRIORITY_UPGRADE_SUCCCESSED)
                if SOURCES_UPDATE_NAME in upgrade_content:
                    self.dbusController.make_properties_restart()
                if BACKEND_PKG_NAME in upgrade_content:
                    if FRONTEND_PKG_NAME not in upgrade_content:
                        self.configs_uncover.setValue("SystemStatus","backend_restart",str(True))
                        if SOURCES_UPDATE_NAME in upgrade_content:
                            self.configs_uncover.setValue("SystemStatus","source_restart",str(True))
                    self.dbusController.Quit(None)
                    return
                if FRONTEND_PKG_NAME not in upgrade_content:
                    if SOURCES_UPDATE_NAME in upgrade_content:
                        self.start_update(fix_env=fix_env)
                    else:
                        self.start_available(fix_env=fix_env)
            else:
                self._update_error_handler(error_code)

        selfupgrade = get_backend(self,InstallBackend.ACTION_SLEF_UPGRADE,InstallBackend.MODE_DEFAULT_STATUS)
        selfupgrade.connect("finished", _selfupgrade_finished)
        selfupgrade.start(upgrade_content)

    def _check_config_upgrade(self,cache,fix_env):
        need_upgrade = False
        self_upgrade = []

        #config包
        for pkg_name in [GROUPS_PKG_NAME]:
            if pkg_name in cache:
                self_pkg = cache[pkg_name]
                if self_pkg.is_installed:
                    if self_pkg.is_upgradable:
                        logging.info("Check: (%s) start upgrading From %s to %s...",pkg_name,\
                                        self_pkg.installed.source_version,self_pkg.candidate.source_version)
                        try:
                            self_pkg.mark_install(True, False, True)
                            self_upgrade.append(pkg_name)
                            need_upgrade = True
                        except SystemError:
                            logging.error("Check: mark %s to upgrade Failed...",pkg_name)
                            raise UpdateBaseError(ERROR_NOT_CONFIGPKG_DEPENDENCIES)
                    else:
                        # 已安装版本
                        cur_version = getattr(self_pkg.installed, "version", None)
                        sources_version = None
                        # 源中最高版本
                        for version_obj in self_pkg.versions:
                            if version_obj.downloadable == False:
                                continue
                            sources_version = version_obj.version
                            self_pkg.candidate = version_obj
                            break

                        if cur_version and sources_version:
                            # 当前版本大于源中的版本 将config包进行降级 
                            if apt_pkg.version_compare(cur_version,sources_version) > 0:
                                logging.info("Check: (%s) need to downgrade from %s to %s...",pkg_name,cur_version,sources_version)
                                self_pkg.mark_install(True, False, True)
                                self_upgrade.append(pkg_name)
                                need_upgrade = True
                        logging.info("Check: (%s:%s) No need to upgrade...",pkg_name,self_pkg.installed.source_version)
                else:
                    logging.info("Check: (%s) start new installing...",pkg_name)
                    try:
                        self_pkg.mark_install(True, False, True)
                        self_upgrade.append(pkg_name)
                        need_upgrade = True
                    except SystemError:
                        logging.error("Check: mark %s to install Failed...",pkg_name)
                        raise UpdateBaseError(ERROR_NOT_CONFIGPKG_DEPENDENCIES)
            else:
                logging.error("Check: groups JSON ConfigPkgs(%s) is not in Cache...",pkg_name)
                raise UpdateBaseError(ERROR_NOT_GROUPS_CONFIG)

        if need_upgrade == True:
            self.dbusController.UpdateDetectStatusChanged(95,_("Priority Upgrade Package being updated"))
            self.start_selfupgrade(self_upgrade,fix_env)
            raise UpdateProgressExit()

    #进行清空所有下载的文件
    def start_clean(self):
        clean_backend = get_backend(self, InstallBackend.ACTION_CLEAN)
        clean_backend.start()

    #进行 dpkg --configure
    def start_fix_incomplete(self):
        #安装回调
        def _fix_finished(parent,success,error_code):
            if success:
                self.start_available()
            else:
                self._update_error_handler(error_code)

        fix_backend = get_backend(self, InstallBackend.ACTION_FIX_INCOMPLETE)
        fix_backend.connect("finished", _fix_finished)
        fix_backend.start()

    # private = True 代表 系统更新的私有配置 False：走本地系统的cache
    # fresh_cache True 代表强制刷新cache False：代表自动检查刷新cache 
    def refresh_cache(self,private=False,fresh_cache=False):
        SOURCES_LIST_FLAG = "sources.list"
        def _reload_private_options():
            if os.path.exists(APT_PRIVATE_PATH):
                apt_pkg.read_config_file(apt_pkg.config, APT_PRIVATE_PATH)
            apt_pkg.init_config()
        def _reload_standard_options():
            if os.path.exists(APT_STANDARD_PATH):
                apt_pkg.read_config_file(apt_pkg.config, APT_STANDARD_PATH)
            apt_pkg.init_config()

        try:
            sources_path = apt_pkg.config["Dir::Etc::SourceList"]
            if private:
                # 检查和加载的配置不同时 重新导入配置 以及加载cache
                if sources_path == SOURCES_LIST_FLAG:
                    _reload_private_options()
                    fresh_cache = True
            else: 
                if sources_path != SOURCES_LIST_FLAG:
                    _reload_standard_options()
                    fresh_cache = True
            if self.cache is None:
                self.cache = MyCache(None)
            elif fresh_cache:
                self.cache.open(None)
                self.cache._initDepCache()
                logging.info("Finished: refreshing Cache for System Update Successfully...")
        except AssertionError:
            raise UpdateBaseError(ERROR_SOFTWARE_INDEX_RROKEN)
        except SystemError as e:
            logging.error(str(e))
            raise UpdateBaseError(ERROR_NOT_INIT_PACKAGESINFIO)


    def _check_network(self):
        self.dbusController.UpdateDetectStatusChanged(5,_("Checking network connection"))
        try:
            obj = self.bus.get_object("org.freedesktop.NetworkManager","/org/freedesktop/NetworkManager")
            interface = dbus.Interface(obj, "org.freedesktop.DBus.Properties")
            retval = interface.Get("org.freedesktop.NetworkManager","Connectivity",timeout=0.5)
        except Exception:
            retval = 4
        #1 表示没有网卡可以使用
        if retval == 1:
            raise UpdateBaseError(ERROR_NETWORK_FAILED)

    #更新important.list的本次升级的列表
    def _update_important(self,reply_handler,error_handler):
        def _update_important_reply(retval):
            if bool(retval) == False:
                error_handler(ERROR_UPDATE_SOURCE_FAILED)
            else:
                reply_handler()
        def _update_important_error(retval):
            logging.error(str(retval))
            error_handler(ERROR_UPDATE_SOURCE_FAILED)
        
        try:
            obj = self.bus.get_object('com.kylin.software.properties', '/com/kylin/software/properties')

            methods = dbus.Interface(obj, dbus_interface='org.freedesktop.DBus.Introspectable').Introspect()
            interface = dbus.Interface(obj, dbus_interface='com.kylin.software.properties.interface')
            if "checkUpdatesInStages" in methods:
                self.dbusController.UpdateDetectStatusChanged(10,_("Updating Source Template")+"(checkUpdatesInStages)")
                interface.checkUpdatesInStages(timeout=20,reply_handler=_update_important_reply,error_handler=_update_important_error)
            else:
                self.dbusController.UpdateDetectStatusChanged(10,_("Updating Source Template")+"(updateSourceTemplate)")
                interface.updateSourceTemplate(timeout=20,reply_handler=_update_important_reply,error_handler=_update_important_error)
        except Exception as e:
            logging.warning(str(e))
            raise UpdateBaseError(ERROR_UPDATE_SOURCE_FAILED)

    def _refresh_uuid(self,success,error_code):
        # 更新完成之后刷新uuid 为了源更新管理器
        if success:
            self.configs_uncover.setValue("statusForProperties","last_install",INSTALL_SUCCESS_FLAG)
        else:
            self.configs_uncover.setValue("statusForProperties","last_install",get_error_num_from_enum(error_code))
        tmpconfig = UpgradeConfig(datadir = "/var/lib/kylin-software-properties/config/", name = "updateID.conf")
        uuid = tmpconfig.getWithDefault("update", "uuid", "")
        self.configs_uncover.setValue("statusForProperties","uuid",uuid)
        
    def _record_finish_status(self,success,install_content,error_code):
        if success:
            self.configs_cover.setValue("statusForFinishPrompt","finished_status",INSTALL_SUCCESS_FLAG)
            self.configs_uncover.setValue("LastUpdateInfo","finished_status",INSTALL_SUCCESS_FLAG)
        else:
            self.configs_cover.setValue("statusForFinishPrompt","finished_status",get_error_num_from_enum(error_code))
            self.configs_uncover.setValue("LastUpdateInfo","finished_status",get_error_num_from_enum(error_code))

        update_version = json_config().getWithDefault("version",default = '')
        self.configs_uncover.setValue("LastUpdateInfo","update_version",update_version)
        self.configs_uncover.setListValue("LastUpdateInfo","update_content",install_content)

    def _update_last_install(self,push_content,upgradeable_push):
        try:
            last_status = self.configs_uncover.getWithDefault("LastUpdateInfo","finished_status",INSTALL_SUCCESS_FLAG)

            # 上次安装成功 不做任何处理
            if last_status == INSTALL_SUCCESS_FLAG or last_status == '':
                pass
            else:
                last_version = self.configs_uncover.getWithDefault("LastUpdateInfo","update_version",'')
                last_content = self.configs_uncover.getlist("LastUpdateInfo","update_content")
                now_version = json_config().getWithDefault("version",default = '')
                
                # 比对版本号不同或者更新内容不相同 则认为认为存在新的推送
                # 推送更新的内容与上次不同
                # 系统为最新时也刷新
                if now_version != last_version or last_content != push_content or \
                                                    push_content != [] and upgradeable_push == []:
                    self.configs_uncover.setValue("LastUpdateInfo","finished_status",INSTALL_SUCCESS_FLAG)
                    self.configs_uncover.setValue("LastUpdateInfo","update_version",'')
                    self.configs_uncover.setListValue("LastUpdateInfo","update_content",[])

                    self.configs_uncover.setValue("statusForProperties","last_install",INSTALL_SUCCESS_FLAG)
                    tmpconfig = UpgradeConfig(datadir = "/var/lib/kylin-software-properties/config/", name = "updateID.conf")
                    uuid = tmpconfig.getWithDefault("update", "uuid", "")
                    self.configs_uncover.setValue("statusForProperties","uuid",uuid)
            
            # 在推送内容不为空，可升级内容为空时 更新版本号
            if push_content != [] and upgradeable_push == []:
                self.sqlite3_server._refresh_system_version()
        except Exception as e:
            logging.error(str(e))

    #更新完成刷新本地源信息
    def _refresh_local_sources(self):
        try:
            logging.info("Now start to fresh local sources(from /var/lib/kylin-system-updater to /etc/apt/sources.list)...")
            obj = self.bus.get_object('com.kylin.software.properties', '/com/kylin/software/properties')
            interface = dbus.Interface(obj, dbus_interface='com.kylin.software.properties.interface')
            interface.replaceUpgradeSource(timeout=0.5)
        except Exception as e:
            logging.warning(str(e))

    def _update_error_handler(self,error_code):
        self.dbusController.UpdateDetectFinishedNext(False,[''],error_code)

    def _get_important_data(self):
        try:
            important_list = []
            if self.is_openkylin == False:
                with open(IMPORTANT_LIST_PATH, 'r') as f:
                    data = f.read()
                    important_list = data.split()
            else:
                important_list = [SYSTEM_UPDATE_GROUPS]

            #去除重复内容
            important_list = list(set(important_list))

            return important_list
        except Exception as e:
            logging.error(str(e))
            return []


    def umount_offline_source(self):
        if os.path.exists("/media/刻录"):
            logging.info("Staring umount /media/刻录")
            p = subprocess.run(["umount","-l","/media/刻录"],timeout=1)
            if p.returncode == 0:
                logging.info("umount offline source to success...")
                return False
            else:
                logging.info("umount offline source to failed...")
                return False

        if is_offline_source():
            if self.now_working != InstallBackend.ACTION_DEFUALT_STATUS:
                # logging.warning('In the process of other task(%d) and Start polling the umount...',self.now_working)
                return True
            try:
                logging.info("Staring umount /media/OfflineSource/")
                p = subprocess.run(["umount","-l","/media/OfflineSource/"],timeout=1)
                if p.returncode == 0:
                    logging.info("umount offline source to success...")
                    return False
                else:
                    logging.info("umount offline source to failed...")
                    return False
            except Exception as exc:
                logging.error(exc)
                return False
        else:
            logging.info("The offline source is not mounted, it does not need to be umounted...")
            return False

    def is_shutdown_install(self):
        if self.task_queue:
            self.last_install_mode = self.task_queue[-1].shutdown_install
            return self.last_install_mode
        else:
            return self.last_install_mode

    def check_shutdown_install(self):
        if self.task_queue and self.task_queue[-1].available and self.task_queue[-1].shutdown_install:
            logging.info("Requires a Shutdown installtion...")
            return 1
        else:
            logging.info("Do not need to perform a shutdown installtion...")
            return 0

    def _check_failed_packages(self,cache):
        need_upgrade = False
        self_upgrade = []
        failed_packages = []

        for pkg in cache:
            if pkg.is_installed and pkg._pkg.current_state == apt_pkg.CURSTATE_HALF_CONFIGURED:
                failed_packages.append(pkg.name)
        
        logging.info("Check: Packages state and Exception pkgs:%r...",failed_packages)

        for pkg_name in failed_packages:
            self_pkg = cache[pkg_name]
            if self_pkg.is_installed and self_pkg.is_upgradable:
                try:
                    logging.info("Check: (%s) start fixing system From %s to %s...",pkg_name,\
                                    self_pkg.installed.source_version,self_pkg.candidate.source_version)
                    self_pkg.mark_install(True,False,True)
                    self_upgrade.append(pkg_name)
                    need_upgrade = True
                except SystemError:
                    logging.error("Check: mark %s to upgrade Failed...",pkg_name)
                    raise UpdateBaseError(ERROR_NOT_FIX_SYSTEM)
            else:
                logging.info("Check: (%s) Not to be installed or upgrade...",pkg_name)
                raise UpdateBaseError(ERROR_NOT_FIX_SYSTEM)
        
        if need_upgrade == True:
            self.dbusController.UpdateDetectStatusChanged(95,_("Repairing the system APT environment"))
            self.start_selfupgrade(push_content=self_upgrade)
            raise UpdateProgressExit()

    def _check_self_upgrade(self,cache,fix_env):
        need_upgrade = False
        self_upgrade = []

        upgrade_list =  [BACKEND_PKG_NAME]

        for pkg_name in upgrade_list:
            if pkg_name in cache:
                self_pkg = cache[pkg_name]
                if self_pkg.is_installed:
                    if self_pkg.is_upgradable:
                        logging.info("Check: (%s) will upgrading From %s to %s...",pkg_name,\
                                        self_pkg.installed.source_version,self_pkg.candidate.source_version)
                        try:
                            logging.info("Check: (%s) start upgrading From %s to %s...",pkg_name,\
                                            self_pkg.installed.source_version,self_pkg.candidate.source_version)
                            self_pkg.mark_install(True,False,True)
                            self_upgrade.append(pkg_name)
                            need_upgrade = True
                        except SystemError:
                            logging.error("Check: mark %s to upgrade Failed...",pkg_name)
                            raise UpdateBaseError(ERROR_NOT_SELFPKG_DEPENDENCIES)
                    else:
                        logging.info("Check: (%s:%s) No need to upgrade...",pkg_name,self_pkg.installed.source_version)
                else:
                    logging.info("Check: (%s) Not to be installed...",pkg_name)
            else:
                logging.error("Check: (%s) The upgrade package is not in Cache...",pkg_name)
        
        if need_upgrade == True:
            if fix_env:
                self.dbusController.UpdateDetectStatusChanged(85,_("Priority Upgrade Package being updated"))
                self.start_selfupgrade(self_upgrade,fix_env)
                raise UpdateProgressExit()

    def _check_other_upgrade(self,cache,fix_env):
        need_upgrade = False
        self_upgrade = []
        
        channel_config = json_config().getWithDefault("update_channel_upgrade",default = None)
        if channel_config == None:
            logging.warning("Json: update_channel_upgrade item is None...")
            upgrade_list =  [BACKEND_PKG_NAME,APTD_PKG_NAME,FRONTEND_PKG_NAME]
        else:
            upgrade_list = channel_config.get("upgrade_list",[BACKEND_PKG_NAME,APTD_PKG_NAME,FRONTEND_PKG_NAME])

        upgrade_list.remove(BACKEND_PKG_NAME)

        for pkg_name in upgrade_list:
            if pkg_name in cache:
                self_pkg = cache[pkg_name]
                if self_pkg.is_installed:
                    if self_pkg.is_upgradable:
                        logging.info("Check: (%s) will upgrading From %s to %s...",pkg_name,\
                                        self_pkg.installed.source_version,self_pkg.candidate.source_version)
                        try:
                            logging.info("Check: (%s) start upgrading From %s to %s...",pkg_name,\
                                            self_pkg.installed.source_version,self_pkg.candidate.source_version)
                            self_pkg.mark_install(True,False,True)
                            self_upgrade.append(pkg_name)
                            need_upgrade = True
                        except SystemError:
                            logging.error("Check: mark %s to upgrade Failed...",pkg_name)
                            raise UpdateBaseError(ERROR_NOT_SELFPKG_DEPENDENCIES)
                    else:
                        logging.info("Check: (%s:%s) No need to upgrade...",pkg_name,self_pkg.installed.source_version)
                else:
                    logging.info("Check: (%s) Not to be installed...",pkg_name)
            else:
                logging.error("Check: (%s) The upgrade package is not in Cache...",pkg_name)
        
        if need_upgrade == True:
            if fix_env:
                self.dbusController.UpdateDetectStatusChanged(95,_("Priority Upgrade Package being updated"))
                self.start_selfupgrade(self_upgrade,fix_env)
                raise UpdateProgressExit()
            
    def _check_push_state(self,update_list):
        sanity_sources = {}
        sanity_system = {}
        packages_database = {}
        desc = ''
        try:
            with open(PACKAGES_DATABASE_PATH,'r') as f:
                packages_database = json.load(f)
        except Exception as exc:
            logging.warning(str(exc))

        try:
            with open(STANDARD_PACKAGES_PATH,'r') as f:
                sanity_system = json.load(f)
        except Exception as exc:
            logging.warning(str(exc))

        try:
            with open(STANDARD_SOURCES_PATH,'r') as f:
                sanity_sources = json.load(f)
        except Exception as exc:
            logging.warning(str(exc))

        # 停止检测功能
        if sanity_system.get("stop_version_match",False):
            return True,desc

        if packages_database and sanity_system and sanity_sources:
            return update_list.check_sanity_verison(packages_database,sanity_system,sanity_sources)

        return True,desc

    def _check_system_incomplete(self,cache,fix_env):
        if cache.dpkgJournalDirty() == True:
            if fix_env:
                self.start_fix_incomplete()
                raise UpdateProgressExit()
            else:
                raise UpdateBaseError(ERROR_NOT_FIX_SYSTEM)

    def _check_system_broken(self,cache,fix_env):
        if cache.is_system_broken() == True:
            if fix_env:
                _,_,pkgs_remove = cache.fix_broken_for_update()
                if pkgs_remove:
                    self.dbusController.UpdateFixBrokenStatus(True,True,pkgs_remove,[''],[''],'','')
                else:
                    self.start_fix_broken()
                raise UpdateProgressExit()
            else:
                raise UpdateBaseError(ERROR_NOT_FIX_SYSTEM)
    
    def _check_system_library(self):
        illegal_library = []

        try:
            system_library = []
            for root, dirs, files in os.walk("/usr/local/lib"):
                for file in files:
                    system_library.append(os.path.join(root, file))
            illegal_library = self._check_necessary_library(system_library)
        except Exception as exc:
            logging.warning(str(exc))

        if len(illegal_library) != 0:
            logging.debug("illegal_library: %s .", illegal_library)
            raise UpdateBaseError(ERROR_NOT_NECESSARY_LIB)

    def _check_necessary_library(self,system_library):
        illegal_library = []
        local_library = {}

        for sl in system_library:
            if ".so." in sl or sl.endswith(".so"):
                if not os.path.islink(sl):
                    process = subprocess.Popen("dpkg -S " + os.path.basename(sl), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    output, error = process.communicate()
                    if not error:
                        install_local, library_legal, library_path = self._check_library_legal(sl, output.decode('utf-8'))
                        if not install_local and not library_legal:
                            illegal_library.append(library_path)
                            local_library[str(sl)] = str(library_path)
                    else:
                        logging.warning("dpkg -S result: %s .",str(error.decode('utf-8')).strip())
                else:
                    logging.debug("'%s' is link file, should ignore.", sl)

        logging.warning("local_library: %s .", local_library)
        return illegal_library

    def _check_library_legal(self, system_library, output):
        is_library_legal = True
        library_legal = ''
        lines = output.splitlines()

        for line in lines:
            if ":" in line:
                lib_path = line.split(":")[1].strip()

                if lib_path == system_library:
                    logging.debug("'%s' is install to local, should ignore.", lib_path)
                    return True,is_library_legal,''

                if os.path.basename(system_library) == os.path.basename(line) and lib_path.startswith("/usr/lib/"):
                    logging.debug("'%s' is legal library.", lib_path)
                    is_library_legal = False
                    library_legal = line
        
        return False,is_library_legal,library_legal
        

    def _setup_dbus(self,bus):
        # check if there is another g-a-i already and if not setup one
        # listening on dbus
        try:
            bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                            bus,do_not_queue=True)
            logging.info("Initiate dbus success ...")
            UpdateManagerDbusControllerUtils(self,bus,bus_name)
            return  UpdateManagerDbusController(self,bus,bus_name)
        except dbus.exceptions.NameExistsException:
            if self.options.replace is False:
                logging.critical("Another daemon is already running")
                sys.exit(1)
            logging.warning("Replacing already running daemon")
            
            retry_reboot_times = 0
            the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                            UPDATER_DBUS_PATH)
            the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                timeout=300)
            time.sleep(1)
            while True:
                retry_reboot_times = retry_reboot_times + 1
                #当重试次数超过5次时退出程序
                if retry_reboot_times > 5:
                    logging.critical("Reboot backend is Failed...")
                    sys.exit(1)
                try:
                    bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                                    bus,
                                                    do_not_queue=True)
                    logging.warning("Replacing already running daemon to Success...")
                    return  UpdateManagerDbusController(self,bus,bus_name)
                except dbus.exceptions.NameExistsException:
                    the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                                    UPDATER_DBUS_PATH)
                    the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                        timeout=300)
                    logging.error("Dbus has not withdrawn and retry reboot times:%d...",retry_reboot_times)
                    time.sleep(1)
    
    def _setup_dbus_utils(self,bus):
        # check if there is another g-a-i already and if not setup one
        # listening on dbus
        try:
            bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                            bus,
                                            do_not_queue=True)
            return  UpdateManagerDbusControllerUtils(self,bus,bus_name)
        except dbus.exceptions.NameExistsException:
            if self.options.replace is False:
                logging.critical("Another daemon is already running")
                sys.exit(1)
            logging.warning("Replacing already running daemon")
            
            retry_reboot_times = 0
            the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                            UPDATER_DBUS_PATH)
            the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                timeout=300)
            time.sleep(1)
            while True:
                retry_reboot_times = retry_reboot_times + 1
                #当重试次数超过5次时退出程序
                if retry_reboot_times > 5:
                    logging.critical("Reboot backend is Failed...")
                    sys.exit(1)
                try:
                    bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                                    bus,
                                                    do_not_queue=True)
                    logging.warning("Replacing already running daemon to Success...")
                    return  UpdateManagerDbusController(self,bus,bus_name)
                except dbus.exceptions.NameExistsException:
                    the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                                    UPDATER_DBUS_PATH)
                    the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                        timeout=300)
                    logging.error("Dbus has not withdrawn and retry reboot times:%d...",retry_reboot_times)
                    time.sleep(1)

    def start_back_upgrade(self, pkglist, mode=InstallBackend.MODE_OTHER):
        try:
            install_backend = get_backend(self, InstallBackend.ACTION_BACKGROUND_UPGRADE, mode)
            install_backend.start_alone(push_content = pkglist)
        except Exception as e:
            logging.error(str(e))

    def start_deb_resolver(self, pkglist = []):
        try:
            install_backend = get_backend(self, InstallBackend.ACTION_INSTALL_DEB_RESOLVER)
            install_backend.start_alone(push_content = pkglist)
        except Exception as e:
            logging.error(str(e))

    # 进行本地deb包安装的操作
    # _check_local_dep : 是否查询本地依赖
    # _auto_satisfy    : 是否通过网络下载依赖
    def start_deb_install(self, deb_path = "", _check_local_dep = False, _auto_satisfy = False, source = '', sender=None):
        header = ''
        desc = ''
        absolute_path, debname = os.path.split(deb_path)
        self.sqlite3_server.deb_metadata.update({"current_install_debfile":deb_path})
        self.sqlite3_server.deb_metadata.update({"absolute_path":absolute_path})
        self.sqlite3_server.deb_metadata.update({"source":source})
        
        self.deb_obj = {}
        UpdateMsg = {}
        try:
            if not os.path.isfile(deb_path):
                logging.info("No such file or directory: %s.",deb_path)
                self.dbusController.InstalldebFinished(False,'No such file or directory .','')
                self.dbusController.InstalldebFinishedWithErrCode(False,'No such file or directory .','',get_error_num_from_enum(ERROR_READ_LOCAL_DEB))
                return
            # 验签提权
            sender_name = get_proc_from_dbus_name(sender)
            caller = get_caller_from_enum(sender_name)
            caller_trans = get_source_name_from_enum(sender_name)

            if not deb_verify(deb_path): #验签失败，提权
                if not self.sqlite3_server.deb_policy_keep:
                    (status,error_string) = PolicyKit_Authority(caller_trans+_(" requires authentication to install software packages."),
                                                sender = sender, InstPolicy = True,
                                                authentication = self.configs_uncover.getWithDefault("InstallAndPurge","install_authority",True),
                                                source=source)
                    if not status:
                        self.dbusController.InstalldebFinished(False,error_string,'')
                        self.dbusController.InstalldebFinishedWithErrCode(False,error_string,'',get_error_num_from_enum(ERROR_INSTALL_DEB_VERIFY))
                        return
                    else:
                        logging.info("Start check deb policy timeout...")
                        
                        if error_string == "USI-policy-close" or error_string == "USI-policy-deter":
                            self.sqlite3_server.deb_policy_keep = False
                            self.sqlite3_server.deb_policy_timestamp = 0
                        else:
                            self.sqlite3_server.deb_policy_keep = True
                            def _check_deb_policy():
                                if self.sqlite3_server.deb_policy_timestamp % 10 == 0:
                                    logging.info("Checking for deb policy timeout(%d)...",self.sqlite3_server.deb_policy_timestamp)
                                if (self.sqlite3_server.deb_policy_timestamp <= 0):
                                    logging.warning("Deb policy timeout")
                                    self.sqlite3_server.deb_policy_keep = False
                                    return False
                                else:
                                    self.sqlite3_server.deb_policy_timestamp = self.sqlite3_server.deb_policy_timestamp - 1
                                return True

                            from gi.repository import GLib
                            self.sqlite3_server.deb_policy_timestamp = 60 * 5
                            GLib.timeout_add_seconds(1,_check_deb_policy)
                
                else:
                    self.sqlite3_server.deb_policy_timestamp = 60 * 5
                    logging.info("Deb policy keep, ignore...")

            self.deb_obj.update({"debname":str(debname)})
            self.deb_obj.update({"old_version":""})
            self.deb_obj.update({"source":str(caller)})
            UpdateMsg.update({"source":str(self.deb_obj.get("source","kylin-system-updater"))})
            
            self.refresh_cache(False,True)
            if self.cache is None:
                deb_cache = MyCache(None)
            else:
                deb_cache = self.cache

            ins = self._suit_install_mode(deb_path, deb_cache)
            UpdateMsg.update({"appname":str(self.debName)})
            self.sqlite3_server.deb_metadata.update({"debname":str(self.debName)})
            self.sqlite3_server.deb_metadata['deblist'].append(str(self.debName))
            UpdateMsg.update({"new_version":str(self.debVersion)})

            status,error_string = ApplicationSecInstall(str(self.debName))
            if not status:
                self.dbusController.InstalldebFinished(False,error_string,'')
                self.dbusController.InstalldebFinishedWithErrCode(False,error_string,'',get_error_num_from_enum(ERROR_INSTALL_DEB_SECURITY))
                return 

            if self._is_broken > 0 or not self.cacheSatisfy:
            # 走 dpkg 安装流程，说明本地apt环境已经损坏 or dep not satisfied
                dep_satisfy, header, desc = self._deb_install(deb_cache, deb_path, _check_local_dep, ins)
                if dep_satisfy:
                    self.dbusController.InstalldebFinished(True, header, desc)
                    self.dbusController.InstalldebFinishedWithErrCode(True, header, desc,get_error_num_from_enum(SUCCESS_INSTALL_DEB))
                    UpdateMsg.update({"status":"success"})
                    UpdateMsg.update({"errorCode":" "})
                else:
                    self.dbusController.InstalldebFinished(False, header, desc)
                    self.dbusController.InstalldebFinishedWithErrCode(False, header, desc,get_error_num_from_enum(ERROR_INSTALL_DEB_DPKG))
                    UpdateMsg.update({"status":"failed"})
                    UpdateMsg.update({"errorCode":str(header+"， "+desc)})
                #dpkg发送数据
                if self.configs_uncover.getWithDefault("SystemStatus", "upload_installer_log", False) == True:
                    UpdateMsg.update({"old_version":str(self.deb_obj.get("old_version","None"))})
                    self.collector.Upgrade_Process_Msg(InstallBackend.ACTION_INSTALL_DEB, UpdateMsg.copy())
                    self.deb_obj = {}
            else:
            # apt 安装流程
                dep_satisfy, header, desc = self._attempt_depends(deb_cache, deb_path, _check_local_dep,_auto_satisfy, ins)
                if dep_satisfy:
                    install_backend = get_backend(self, InstallBackend.ACTION_INSTALL_DEB)
                    logging.info("source name: %s.", source)
                    install_backend.start_alone(push_content=deb_path,_is_install=_auto_satisfy,caller=source)
                else:
                    self.dbusController.InstalldebFinished(False, header, desc)
                    self.dbusController.InstalldebFinishedWithErrCode(False, header, desc,get_error_num_from_enum(ERROR_INSTALL_DEB_DEPEND))
        except UpdateBaseError as excep:
            self.dbusController.InstalldebFinished(False,excep.header,excep.desc)
            self.dbusController.InstalldebFinishedWithErrCode(False,excep.header,excep.desc,get_error_num_from_enum(excep.code))
            UpdateMsg.update({"old_version":str(self.deb_obj.get("old_version","None"))})
            UpdateMsg.update({"errorCode":str(excep.header+"， "+excep.desc)})
            UpdateMsg.update({"status":"failed"})
            self.collector.Upgrade_Process_Msg(InstallBackend.ACTION_INSTALL_DEB, UpdateMsg.copy())
        except Exception as e:
            logging.info(str(e))
            tbk = traceback.format_exc()
            logging.error(str(tbk))
            self.dbusController.InstalldebFinished(False, str(e), desc)
            self.dbusController.InstalldebFinishedWithErrCode(False,str(e),desc,get_error_num_from_enum(ERROR_INSTALL_DEB_OTHER))
            UpdateMsg.update({"old_version":str(self.deb_obj.get("old_version","None"))})
            UpdateMsg.update({"errorCode":str(e)})
            UpdateMsg.update({"status":"failed"})
            self.collector.Upgrade_Process_Msg(InstallBackend.ACTION_INSTALL_DEB, UpdateMsg.copy())

    def _system_reboot_shutdown(self,status="reboot",prohibit_shutdown=False):
        if status == "shutdown":
            args = ["/sbin/shutdown","-h","now"]
        else:
            args = ["/sbin/shutdown","-r","now"]

        if prohibit_shutdown:
            pass
        else:
            logging.info("The device will be %s by run system command(%r)...",status,args)
            p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
            if p.returncode != 0:
                logging.info("The device has (%s) Failed...",status)
                return False
    #进行删除的操作
    def start_purge_pkgs(self,pkgs_list):
        try:
            # 是否有破损的包
            self.refresh_cache(False,True)
            if self.cache is None:
                deb_cache = MyCache(None)
            else:
                deb_cache = self.cache
            broken_count = deb_cache._depcache.broken_count
            if broken_count > 0:
                _success,header,desc = self._dpkg_purge_pkgs(pkgs_list)
                if _success == True:
                    logging.info(header)
                    self.dbusController.PurgePackagesFinished(_success,'',desc)
                    self.dbusController.PurgePackagesFinishedWithPkgname(_success,'',desc,",".join(pkgs_list))
                    self.dbusController.PurgePackagesFinishedWithErrCode(_success,'',desc,",".join(pkgs_list),get_error_num_from_enum(SUCCESS_PURGE_PKG))
                else:
                    self.dbusController.PurgePackagesFinished(_success,header,desc)
                    self.dbusController.PurgePackagesFinishedWithPkgname(_success,header,desc,",".join(pkgs_list))
                    self.dbusController.PurgePackagesFinishedWithErrCode(_success,header,desc,",".join(pkgs_list),get_error_num_from_enum(ERROR_PURGE_PKG_DPKG))
            else:
                purge_backend = get_backend(self, InstallBackend.ACTION_REMOVE_PACKAGES)
                purge_backend.start_alone(push_content = pkgs_list)
            
            deb_cache.close()
        except Exception as e:
            logging.error(e)
            tbk = traceback.format_exc()
            logging.error(str(tbk))

    def _dpkg_purge_pkgs(self,pkgs_list):
        success = False
        pkg_list_str = " ".join(pkgs_list)
        args = ["dpkg", "-r"]
        args.extend([pkg_list_str])

        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        success = p.returncode == 0
        return success,p.stdout,''

    def check_config_aptdeamon(self):
    #检查是否需要重新启动aptdeamon 目前需要重启的有限速功能
        if self.init_config_aptdeamon == True:
            self.init_config_aptdeamon = False
            self.dbusController.set_aptdeamon_environ("init","config")

    # 是否查找本地依赖
    def _attempt_depends(self,deb_cache, deb_path,_check_local_dep,_auto_satisfy, _install):  
        depends_list = []
        depends_pkg = []
        satisfy_list = []
        depends_count = 0
        _local_satisfy = False
        error_string = ''
        error_desc = ''
        absolute_path, debname = os.path.split(deb_path)
        # 依赖不满足的情况
        if len(_install) > 0:
            if _check_local_dep: #查找本地
                # 需要查找本地依赖
                if len(_install) > 0:
                    for pkg in deb_cache:
                        if pkg.marked_install and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                        elif pkg.marked_upgrade and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                if len(depends_pkg)>0: #查找本地deb包
                    depends_list = [debfile for debfile in os.listdir(absolute_path) if debfile.endswith(".deb")]
                    for depends in depends_pkg:
                        for debfile in depends_list:
                            if depends.name in debfile and depends.candidate.version in debfile:
                                depends_count += 1
                                satisfy_list.append(debfile)
                    if depends_count < len(depends_pkg):
                        #本地依赖不满足 判断源下载
                        if _auto_satisfy:
                            _local_satisfy = True
                        elif not _auto_satisfy:
                            _local_satisfy = False
                            error_string = str(debname.split("_")[0])+_("dependency is not satisfied")
                            error_desc   = ",".join(_install)
                            logging.error(error_string+ error_desc)
                    else:
                        #将应用包与依赖包拷贝至archive目录安装
                        try:
                            if debname not in satisfy_list:
                                satisfy_list.append(debname)
                            for satisfy in satisfy_list:
                                shutil.copy(os.path.join(absolute_path,satisfy),"/var/cache/apt/archives/")
                                logging.info("move debpkg: %s",satisfy)
                            _local_satisfy = True
                        except Exception as e:
                            logging.info(str(e))
                return _local_satisfy,error_string,error_desc
            elif not _check_local_dep and _auto_satisfy:
                _local_satisfy = True
                if _install:
                    error_string = str(debname.split("_")[0])+_("dependency is not satisfied will download")
                    error_desc   = ",".join(_install)
                    logging.error(error_string+error_desc)
                    return _local_satisfy,error_string,error_desc
            elif not _check_local_dep and not _auto_satisfy:
                _local_satisfy = False
                if _install:
                    error_string = str(debname.split("_")[0])+_("dependency is not satisfied")
                    error_desc   = ",".join(_install)
                    logging.error(error_string+error_desc)
                    return _local_satisfy,error_string,error_desc
        # 依赖满足
        else:
            _local_satisfy = True
            error_string = ''
            error_desc=''
            return _local_satisfy,error_string,error_desc
    
    def _deb_install(self, deb_cache, deb_path, _check_local_dep, _install):
        local_depends_deb_list = []
        satisfy_list = []
        noSatisfyList = []
        self.depends_count = 0
        error_string = ''
        header = ''
        desc = ''
        absolute_path, debname = os.path.split(deb_path)
        debPackage = DebPackage(deb_path, deb_cache)
        self.absolute_path = absolute_path

        if _check_local_dep: # 需要查找本地依赖
            if len(self.noSatisfyList) > 0:
                for nS in self.noSatisfyList:
                    if str(nS[0][0]) not in noSatisfyList:
                        noSatisfyList.append(str(nS[0][0]))

            if len(noSatisfyList) > 0: #查找本地deb包
                local_depends_deb_list = [debfile for debfile in os.listdir(absolute_path) if debfile.endswith(".deb")]

                for nS in self.noSatisfyList:
                    for debfile in local_depends_deb_list:
                        # if "%3a" in debfile:
                        #     debfile=debfile.replace("%3a",":")
                        dep = DebPackage(os.path.join(absolute_path,debfile), deb_cache)
                        
                        if str(nS[0][0]) == dep.pkgname:# and debfile not in satisfy_list:
                            is_resolve = self._depend_resolver(dep, debfile, deb_cache, local_depends_deb_list, satisfy_list)
                            if is_resolve and debfile not in satisfy_list:
                                self.depends_count += 1
                                satisfy_list.append(debfile)

                if self.depends_count < len(noSatisfyList): #本地依赖不满足
                    error_string = str(debPackage.pkgname)+_("dependency is not satisfied")+", ".join(noSatisfyList)
                    logging.error(error_string)
                    header = error_string
                    return False,header, desc
                else: #安装依赖包
                    if debname not in satisfy_list:
                        satisfy_list.append(debname)
                    for satisfy in satisfy_list:
                        try:
                            deb = DebPackage(os.path.join(absolute_path,satisfy))
                            # iprogress = LogInstallProgress(os.path.join(absolute_path,satisfy))
                            ret = deb.install() #install_progress=iprogress
                            if ret != 0:
                                return False, "软件包安装失败!", desc
                        except Exception as e:
                            logging.error(e)
                            return False, str(e), desc
                    return True, header, desc
            else:
                try:
                    deb = DebPackage(deb_path)
                    # iprogress = LogInstallProgress(os.path.join(absolute_path,satisfy))
                    ret = deb.install() #install_progress=iprogress
                    if ret != 0:
                        return False, "软件包安装失败!", desc
                except Exception as e:
                    logging.error(e)
                    return False, str(e), desc
                return True, header, desc
        elif not _check_local_dep:
            error_string = str(debPackage.pkgname)+_("dependency is not satisfied")+":".join(_install)
            logging.error(error_string)
            header = error_string
            return False,header, desc

    def _depend_resolver(self, deb, debfile, deb_cache, local_depends_deb_list, satisfy_list):
        depends_cnt = 0
        is_resolve = False
        depends = deb.depends
        noSatisfyList = self._gen_noSatisfyList(depends, deb_cache)
        if len(noSatisfyList) > 0:
            for nS in noSatisfyList:
                for debfile in local_depends_deb_list:
                    if "%3a" in debfile:
                        debfile=debfile.replace("%3a",":")
                    dep = DebPackage(os.path.join(self.absolute_path,debfile), deb_cache)

                    if str(nS[0][0]) == dep.pkgname:
                        is_resolve = self._depend_resolver(dep, debfile, deb_cache, local_depends_deb_list, satisfy_list)
                        if is_resolve and debfile not in satisfy_list:
                            depends_cnt += 1
                            satisfy_list.append(debfile)
            if depends_cnt < len(noSatisfyList):
                is_resolve = False
        else:
                is_resolve = True
        return is_resolve

    def _suit_install_mode(self, deb_path, deb_cache):
        self._is_broken = False
        self.cacheSatisfy = False

        # 检查本地破损
        try:
            broken_count = deb_cache._depcache.broken_count
            logging.info("Install deb_package: check broken: %d. ", broken_count)
            self.debVersion = 0
            debPackage = DebPackage(deb_path, deb_cache)
            debPackage.check(allow_downgrade=True)
            self.debVersion = debPackage._sections["Version"]
            self.debName = debPackage.pkgname
            logging.info("Install deb_package: required changes")
            (install, remove, unauth) = debPackage.required_changes    # need in cach
            if broken_count > 0:
                self._is_broken = True
            else :
                self._is_broken = False
            if debPackage.pkgname in deb_cache:
                pkg = deb_cache[debPackage.pkgname]
                if pkg.is_installed:
                    self.deb_obj.update({"old_version":str(pkg.installed.version)})
        except apt_pkg.Error as e:
            logging.error(str(e))
            if "E:read" in str(e):
                raise UpdateBaseError(ERROR_READ_LOCAL_DEB)
            else:
                raise UpdateBaseError(ERROR_LOCAL_DEB_FORMAT)
        except Exception as e:
            logging.error(str(e))
            raise UpdateBaseError(ERROR_INSTALL_DEB_BASE)
        
        # 不满足的依赖列表
        depends = debPackage.depends
        self.noSatisfyList = self._gen_noSatisfyList(depends, deb_cache)
        _list = []
        # cache是否满足
        if len(install) == 0 and len(self.noSatisfyList) == 0:
            self.cacheSatisfy = True
        else:
            for ns in self.noSatisfyList:
                for or_group in ns:
                    for pkg in install:
                        if pkg == or_group[0]:
                            _list.append(ns)
            if len(_list) == len(self.noSatisfyList):
                self.cacheSatisfy = True
            else:
                self.cacheSatisfy = False
        logging.info("Cache satisfy is %r.",self.cacheSatisfy)
        return install
    
    def _gen_noSatisfyList(self, depends, deb_cache):
        _noSatisfyList = []
        _group_satify=False
        providers = []
        for or_group in depends:
            for deb_info in or_group:
                debname,ver,oper = deb_info
                if ":" in debname:
                    debname=debname.split(":")[0]
                try:
                    if _group_satify:
                        logging.warning("Group depend satify, ignore %s.", debname)
                        continue
                    if debname not in deb_cache:
                        if  deb_cache.is_virtual_package(debname):
                            logging.warning("The cache has no package named '%s', this is a virtual package...", debname)
                            providers = deb_cache.get_providing_packages(debname)
                        # hear!
                        if len(providers) < 1:
                            logging.error("Can not found depend %s.", debname)
                            continue
                        depname = providers[0].name
                        pkg = deb_cache[depname]
                        cand = deb_cache._depcache.get_candidate_ver(pkg._pkg)
                        logging.warning("Found virtual package %s, providers name: %s.", debname,depname)
                        if not cand:
                            _group_satify = False
                        if not apt_pkg.check_dep(cand.ver_str, oper, ver):
                            _group_satify = False
                        else:
                            _group_satify = True
                    else:
                        pkg = deb_cache[debname]
                        if (apt_pkg.check_dep(pkg.candidate.source_version, oper, ver)) or (pkg.installed and apt_pkg.check_dep(pkg.installed.version, oper, ver)) or (not pkg.installed and apt_pkg.check_dep(pkg.candidate.version, oper, ver)) and _group_satify == False:
                            # if not apt_pkg.check_dep(cand, oper, ver): #no candidate
                            _group_satify = True
                except Exception as e:
                    logging.error(str(e))
            if _group_satify == False and or_group not in _noSatisfyList:
                _noSatisfyList.append(or_group)
            _group_satify = False
        return _noSatisfyList


class AptP2pConfigManager():
    APT_P2P_FILE = "/etc/apt-p2p/apt-p2p.conf"
    HEADER_DSC = "apt-p2p config(/etc/apt-p2p/apt-p2p.conf) is not exists..."
    def __init__(self):
        if os.path.exists(self.APT_P2P_FILE):
            self.p2pConfigs = UpgradeConfig(datadir = "/etc/apt-p2p/", name = "apt-p2p.conf")
        else:
            self.p2pConfigs = None
    
    def get_bootstrap(self):
        if self.p2pConfigs == None:
            return self.HEADER_DSC

        return self.p2pConfigs.getWithDefault("apt_p2p_Khashmir", "BOOTSTRAP", "Failed")

    def set_bootstrap(self,value):
        if self.p2pConfigs == None:
            return self.HEADER_DSC
        self.p2pConfigs.setValue("apt_p2p_Khashmir","BOOTSTRAP",str(value),True)

class DelayShutdownLock():
    # 禁止关机锁文件路径
    FILELOCK_PATH = "/tmp/lock/"
    SHUTDOWN_BLOCK_FILELOCK = "kylin-update.lock"
    
    def __init__(self,parent):
        self.inhibit_lock = None
        self.pidfile = None
        self.parent = parent
        self.bus = self.parent.bus

    #安装时禁止关机 进行加锁
    def lock(self, caller='kylin-system-updater',mode='delay'):
        """
        Send a dbus signal to logind to not suspend the system, it will be
        released when the return value drops out of scope
        """
        if self.inhibit_lock == None:
            try:
                from gi.repository import Gio, GLib
                logging.info("Start acquiring Shutdown lock(%s)...",mode)
                connection = Gio.bus_get_sync(Gio.BusType.SYSTEM)
                
                var, fdlist = connection.call_with_unix_fd_list_sync(
                    'org.freedesktop.login1', '/org/freedesktop/login1',
                    'org.freedesktop.login1.Manager', 'Inhibit',
                    GLib.Variant('(ssss)',
                                ('shutdown:sleep',
                                caller, _("Installing or removing software packages."),
                                mode)),
                    None, 0, -1, None, None)
                self.inhibit_lock = Gio.UnixInputStream(fd=fdlist.steal_fds()[var[0]])
                logging.info("Shutdown Has been locked...")
                return True
            except Exception as e:
                logging.error(e)
                return False
        else:
            logging.info("No need to acquired lock and Shutdown Has been locked...")
            return False

    #解锁禁止关机 
    def unlock(self):
        try:
            if self.inhibit_lock != None:
                self.inhibit_lock.close()
                self.inhibit_lock = None
                logging.info("Shutdown Has been unlocked...")
            else:
                logging.info("Not locked and Quitting ...")
        except Exception as e:
            self.inhibit_lock = None
            logging.error("unlock failed." + str(e))

    def connect_logind(self):
        try:
            logind_proxy = self.bus.get_object('org.freedesktop.login1', '/org/freedesktop/login1')
            def prepare_for_shutdown_handler(active):
                """ Handle PrepareForShutdown() """
                if not active:
                    logging.warning("PrepareForShutdown(false) received, "
                                    "this should not happen")

                logging.info("Prepare For Shutdown arrived, starting final iterations....")
                
                device_status = self._check_device_status()

                self._pre_install_preparation()

                self.parent.start_poweroff(device_status)

            logind_proxy.connect_to_signal(
                    "PrepareForShutdown", prepare_for_shutdown_handler)

            self._start_monitor_device()
        except Exception as exc:
            logging.error(exc)

    def _pre_install_preparation(self):
        self._start_plymouth_splash()
        self._stop_lightdm_service()
        self._switch_to_tty1()
        for i in range(3):
            self._stop_logind_service()
            time.sleep(1)
        self.unlock()
        self._switch_to_tty1()

    def _start_plymouth_splash(self):
        try:
            
            if os.path.exists("/sbin/plymouthd"):
                if not self._check_device_type():
                    args = ["/sbin/plymouthd", "--mode=boot","--attach-to-session"]
                else:
                    os.environ["DISPLAY"] = ":0"
                    args = ["/sbin/plymouthd", "--mode=shutdown","--attach-to-session"]
                logging.info("Running plymouth:%r",args)
                subprocess.run(args,timeout=3)
            if os.path.exists("/bin/plymouth"):
                subprocess.Popen(["/bin/plymouth", "show-splash","--wait"])
                subprocess.call(["/bin/plymouth","system-update","--progress=0"])
            
            self._progress_to_plymouth(5)
            mesa = _("The system is installing updates. Please do not turn off your computer. ")
            self._message_to_plymouth(mesa)
        except Exception as exc:
            logging.error(exc)

    def _message_to_plymouth(self,message):
        subprocess.call(["/bin/plymouth", "message", "--text", message])

    def _progress_to_plymouth(self,progress):
        tmp = ("--progress=%d"%progress)
        run_cmd = ["/bin/plymouth","system-update",tmp]
        subprocess.call(run_cmd)

    def _stop_logind_service(self):
        try:
            with open("/usr/bin/touchpad-state", 'w') as file:
                file.write("# fix bug:202730 ,The scripts was cleared by system upgrade.\n")

            p = subprocess.run(["systemctl", "stop","systemd-logind.service"],timeout=1)
            logging.warning("Restart systemd logind service for shutdown install mode..."+str(p.stdout))
        except Exception as exc:
            logging.error(exc)

    def _stop_lightdm_service(self):
        logging.info("Stopping lightdm service")
        try:
            p = subprocess.run(["systemctl","stop","lightdm.service"],timeout=1)
            if p.returncode == 0:
                logging.info("start stop lightdm.service to success...")
                return True
            else:
                logging.info("start stop lightdm.service to failed...")
                return False
        except Exception as exc:
            logging.error(exc)

    def _switch_to_tty1(self):
        logging.info("Force switch to tty1...")
        try:
            p = subprocess.run(["chvt","1"],timeout=0.5)
            if p.returncode == 0:
                logging.info("switch to tty1 successly...")
                return True
            else:
                logging.info("switch to tty1 success to failed...")
                return False
        except Exception as exc:
            logging.error(exc)

    def _check_device_status(self):
        try:
            state = "reboot"
            with open("/tmp/monitor-dbus-upgrade.log", 'r') as f:
                text = f.read()
                if "member=PowerOff" in text:
                    state = "shutdown"
            return state
        except Exception as exc:
            logging.error(exc)
            return "reboot"

    def _start_monitor_device(self): 
        try:
            args = ["/usr/share/kylin-system-updater/monitor-reboot"]
            p = subprocess.run(args,stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
            logging.info("Start monitor device Trigger mothod for Reboot and Poweroff...")
        except Exception as exc:
            logging.error(exc)
    
    def _check_device_type(self):
        try:
            with open("/proc/cpuinfo", 'r') as f:
                for line in f.readlines():
                    if "Hardware" in line and ":" in line:
                        cpuinfos = line.split(':')
                        if "PANGU M900" in cpuinfos[1]:
                            logging.info("Get cpuinfo: PANGU M900")
                            return True
        except Exception as exc:
            logging.error(exc)
        return False