#!/usr/bin/python3

import os
import json
import logging
import apt_pkg
import threading
from apt import Cache
from gettext import gettext as _
from SystemUpdater.Core.errors import *
from SystemUpdater.Core.enums import *
from SystemUpdater.Core.UpdateEssentialItem import UpdateEssentialItem
from SystemUpdater.Core.utils import is_offline_source

class UpdatesAvailable():
    """
    Represent the (potentially partial) results of an unattended-upgrades
    run
    """

    # 更新的缓存数据保存位置 
    UPDATE_AVAILABLE_DATA = "/var/cache/kylin-system-updater/update_available_data.json"

    def __init__(self,parent):
        self.parent = parent
        self.dbus_send = self.parent.dbusController
        self.uuconfigs = self.parent.uuconfigs
        self.configs_cover = self.parent.configs_cover

        self.update_essential = UpdateEssentialItem()
        self.cache = self.parent.cache
        self.last_task = ''
        self.resolver_name = "total_pkgs"

        self.steps_queue = []
        self._problem_resolver = {}
        self.updating_data = {}
        self.upgrade_singles = []
        self.upgrade_groups = []
        self.upgrade_steps = []
        self.device_status = ''
        self.shutdown_install = True
        self.signal_trigger = False
        self.offline_source = False

        self.cloud_env = self.check_cloud_env()
        if self.cloud_env:
            self.shutdown_install = False

        self._available = True 

        if "en_US" in os.environ["LANG"]:
            self.lang = "en_US"
        else:
            self.lang = "zh_CN"

    # push_content = [] 时 全部更新 
    # push_content = ['bao'] 部分更新 
    # 
    def start(self,problem = False,push_content = [],update_data = None):
        total_remove = []

        if not problem:
            self.upgrade_singles,self.upgrade_groups,self.upgrade_steps = update_data.classify_content(push_content)
            self.updating_data = update_data.copy_pkgs_data(self.upgrade_groups+self.upgrade_steps)

            self.shutdown_install = self._check_shutdown_install(self.upgrade_groups+self.upgrade_steps)

            # 适配云系统 没有关机更新界面
            if self.cloud_env:
                self.shutdown_install = False
            
            all_install,all_upgrade = self._get_pkgs_from_data(self.upgrade_groups)

            # self._output_resolver_pkgs(all_install + self.upgrade_singles + all_upgrade)

            pkgs_install,pkgs_upgrade,pkgs_remove = self.cache.make_problem_resolver(self.resolver_name,\
                                                                    all_install+self.upgrade_singles,all_upgrade)
            
            if self.configs_cover.getWithDefault("TestsConfig", "force_dist_upgrade",False):
                raise

            total_remove += pkgs_remove

            self._save_groups_resolver(pkgs_install,pkgs_upgrade,pkgs_remove,\
                            getattr(self.cache, "required_download",0),getattr(self.cache, "additional_required_space",0))
            
            if self.upgrade_steps != []:
                step_remove = self._make_steps_resolver(self.upgrade_steps)
                total_remove += step_remove
        else:
            pkgs_install,pkgs_upgrade,pkgs_remove = self.cache.updateDistUpgrade()

            self._save_groups_resolver(pkgs_install,pkgs_upgrade,pkgs_remove,\
                            getattr(self.cache, "required_download",0),getattr(self.cache, "additional_required_space",0))

            total_remove += pkgs_remove

            # 给全盘更新伪造一个升级内容
            self.upgrade_groups = ["dist-upgrade"]

        self.steps_queue.insert(0,self.resolver_name)

        total_remove = list(set(total_remove))

        logging.info("ProblemResolver of the deletion package list:%s",str(total_remove))                
        self.update_essential.check_white(total_remove)
        
        logging.info("RESOLVER:ALL install:%d , upgrade:%d remove:%d ",len(pkgs_install),len(pkgs_upgrade),\
                                                                    len(total_remove))
        
        is_remove_pkgs = len(total_remove) != 0
        
        delete_des,raw_des= self._supple_description(self.cache,total_remove)

        if not problem:
            self.dbus_send.UpdateDependResloveStatus(True,is_remove_pkgs,total_remove,raw_des,delete_des,'','')
        else:
            self.dbus_send.DistupgradeDependResloveStatus(True,is_remove_pkgs,total_remove,raw_des,delete_des,'','')
        
        if is_remove_pkgs:
            threading_emulate = threading.Thread(target=self._emulate_calcul_delete,args=((pkgs_upgrade + pkgs_install),))
            threading_emulate.start()

        if self.shutdown_install == True:
            self.signal_trigger = self._check_install_trigger(self.cache)

        # 是否使用离线源更新
        self.offline_source = is_offline_source()

    def get_content(self):
        return self.upgrade_singles + self.upgrade_groups

    def _save_groups_resolver(self,pkgs_install,pkgs_upgrade,pkgs_remove,required_download,required_install):
        put_data = {self.resolver_name:{'install_pkgs':pkgs_install,'upgrade_pkgs':pkgs_upgrade,'reomve_pkgs':pkgs_remove,\
                                        'required_download':required_download,'required_install':required_install}}
        self._problem_resolver.update(put_data)

    def _save_steps_resolver(self,step_name,pkgs_install,pkgs_upgrade,pkgs_remove):
        put_data = {step_name:{'install_pkgs':pkgs_install,'upgrade_pkgs':pkgs_upgrade,'reomve_pkgs':pkgs_remove}}
        self._problem_resolver.update(put_data)

    #获取下载大小
    def get_required_download(self):
        return self._problem_resolver.get(self.resolver_name,{}).get('required_download',0)

    #获取安装大小
    def get_required_install(self):
        return self._problem_resolver.get(self.resolver_name,{}).get('required_install',0)

    #获取全部要下载或者安装的软件包
    def get_resolver_upgrade(self):
        pkgs_install = []
        pkgs_upgrade = []
        pkgs_remove = []
        pkgs_install = self._problem_resolver.get(self.resolver_name,{}).get('install_pkgs',[])
        pkgs_upgrade = self._problem_resolver.get(self.resolver_name,{}).get('upgrade_pkgs',[])
        pkgs_remove = self._problem_resolver.get(self.resolver_name,{}).get('reomve_pkgs',[])

        return pkgs_install,pkgs_upgrade,pkgs_remove

    ##获取分步要下载或者安装的软件包
    def get_resolver_steps(self,content):
        pkgs_install = []
        pkgs_upgrade = []
        pkgs_remove = []
        pkgs_install = self._problem_resolver.get(content,{}).get('install_pkgs',[])
        pkgs_upgrade = self._problem_resolver.get(content,{}).get('upgrade_pkgs',[])
        pkgs_remove = self._problem_resolver.get(content,{}).get('reomve_pkgs',[])

        return pkgs_install,pkgs_upgrade,pkgs_remove

    def _supple_description(self,cache,remove_pkgs):
        delete_desc = []
        raw_description = []
        for pkg in remove_pkgs:
            pkg_obj = cache[pkg]
            raw_description.append(getattr(pkg_obj.candidate, "summary", ''))
            delete_desc.append('')

        return delete_desc,raw_description

    def _emulate_calcul_delete(self,total_pkg):
        logging.info("[DELETE_DETAILS] Start calculating delete dependencies...")
        delete_pkgs = []
        last_delete_count = 0
        cache = Cache()

        with cache.actiongroup():
            if cache.get_changes():
                cache.clear()
            for pkg in total_pkg:
                if '#' in pkg:
                    pkg = pkg.split("#")[0]
                elif '=' in pkg:
                    pkg = pkg.split("=")[0]

                try:
                    cache[pkg].mark_install()
                except Exception as e:
                    logging.error(str(e))
                    break

                if cache.delete_count != last_delete_count:
                    last_delete_count = cache.delete_count
                    for pkg_obj in cache.get_changes():
                        if pkg_obj.marked_delete and pkg_obj.name not in delete_pkgs:
                            delete_pkgs.append(pkg_obj.name)
                            delete_pkgs.append(pkg)
                            logging.info("[DELETE_DETAILS] %s Will be deleted Due to Install or Upgrade %s",pkg_obj.name,pkg)
        
        logging.info("[DELETE_DETAILS] End calculating delete dependencies...")
        cache.close()

    def _make_steps_resolver(self,steps):
        total_remove = []
        for step_name in steps:
            step_install,step_upgrade = self._get_pkgs_from_data([step_name])

            step_install,step_upgrade,step_remove = self.cache.make_problem_resolver(step_name,step_install,step_upgrade)

            logging.info("RESOLVER(%s):Steps install:%d , upgrade:%d remove:%d ",step_name,len(step_install),\
                                                                            len(step_upgrade),len(step_remove))

            if len(step_install) == 0 and len(step_upgrade) == 0 and len(step_remove) == 0:
                logging.warning("Steps(%s) This update cannot detect the upgradeable package.",step_name)
                continue

            self._save_steps_resolver(step_name,step_install,step_upgrade,step_remove)

            total_remove += step_remove

            self.steps_queue.insert(0,step_name)
        
        return total_remove

    def _get_pkgs_from_data(self,content):
        pkgs_install = []
        pkgs_upgrade = []

        for cont in content:
            pkgs_install += self.updating_data.get(cont,{}).get('install_list',[])
            pkgs_upgrade += self.updating_data.get(cont,{}).get('upgrade_list',[]) 

        return pkgs_install,pkgs_upgrade

    def check_steps_reboot(self,content):
        return self.updating_data.get(content,{}).get('need_reboot',False)

    def _check_shutdown_install(self,content):
        # 读取策略下发 策略优先级最高
        self.uuconfigs.reRead()
        if self.uuconfigs.getWithDefault("updateStrategiesManager","strategiesState",False) == True:
            install_type = self.uuconfigs.getWithDefault("updateStrategiesManager","installType",DEFAULT_INSTALL_TYPE)
            if install_type == DEFAULT_INSTALL_TYPE:
                pass
            elif install_type == RUNTIME_INSTALL_TYPE:
                return False
            else:
                return True
        
        # 未存在策略时 走默认配置
        for con in content:
            if self.updating_data.get(con,{}).get('install_type',RUNTIME_INSTALL_TYPE) == POWEROFF_INSTALL_TYPE:
                return True        
        return False

    def _check_install_trigger(self,cache):
        pkg_name = "ukui-session-manager"
        targe_version = "3.1.1.10-0k1.2"
        if pkg_name in cache:
            self_pkg = cache[pkg_name]
            if self_pkg.is_installed:
                result = apt_pkg.version_compare(self_pkg.installed.source_version, targe_version)
                if result < 0:
                    logging.info("Check: Shutdown Install Trigger Mode For Reboot Trigger...")
                    return True
                else:
                    logging.info("Check: Shutdown Install Trigger Mode For Dbus Trigger...")
                    return False
        else:
            return False

    def get_post_notify(self):
        notify = self.updating_data.get(self.last_task,{}).get("users_notify",{}).get("post_install_notify",None)
        if notify == None:
            return ''

        return notify.get(self.lang,'')

    def get_mid_notify(self,content):
        notify = self.updating_data.get(content,{}).get("users_notify",{}).get("mid_install_notify",None)
        if notify == None:
            return ''

        return notify.get(self.lang,'')

    def get_pre_script(self):
        script_list = []
        content = ''

        for groups in self.upgrade_groups:
            content = groups
            break

        script = self.updating_data.get(content,{}).get("script",None)
        if script == None:
            return script_list
        
        for scr in ["pre_update_status_check","pre_update_script"]:
            run_path = script.get(scr,[])

            if isinstance(run_path,list) == True:
                for sc in run_path:
                    if sc != '':
                        script_list.append(sc)
            elif isinstance(run_path,str) == True:
                if run_path != '':
                    script_list.append(run_path)
            else:
                pass

        return script_list
        
    def get_download_progress(self):
        if self.shutdown_install == True:
            return 100
        else:
            return 50

    def get_install_progress(self,content):
        #为了 解决获取包列表和获取进度传输的数据不能协同的问题
        if content == self.resolver_name:
            for groups in self.upgrade_groups:
                content = groups
                break
        begin = self.updating_data.get(content,{}).get('progress_begin',0)
        end = self.updating_data.get(content,{}).get('progress_end',0)

        #当计算单包时 他们是没有进度的 补充进度
        if begin == 0 and end == 0:
            if self.shutdown_install == True:
                begin = 0
                end = 100
            else:
                begin = 50
                end = 100

        return begin,end

    def get_post_script(self):
        script_list = []
        content = ''

        for groups in self.upgrade_groups:
            content = groups
            break

        script = self.updating_data.get(content,{}).get("script",None)
        if script == None:
            return script_list
        
        for scr in ["post_update_status_check","post_update_script"]:
            run_path = script.get(scr,[])

            if isinstance(run_path,list) == True:
                for sc in run_path:
                    if sc != '':
                        script_list.append(sc)
            elif isinstance(run_path,str) == True:
                if run_path != '':
                    script_list.append(run_path)
            else:
                pass

        return script_list

    def check_cloud_env(self):
        try:
            with open("/sys/devices/virtual/dmi/id/chassis_vendor", 'r') as f:
                data = f.read()
                data = str(data).strip()
                if data == "Huawei Inc.":
                    return True

            with open("/sys/devices/virtual/dmi/id/chassis_asset_tag", 'r') as f:
                data = f.read()
                data = str(data).strip()
                if data == "HUAWEICLOUD":
                    return True

            return False
        except Exception as e:
            logging.warning(str(e))
            return False

    def _output_resolver_pkgs(self,pkgs):
        OUTPUT_JSON = "/var/lib/kylin-system-updater/json/requires_upgrade_pkgs.json"
        try:
            output_upgrade = {}

            pkg_string = ''
            for pkg in pkgs:
                pkg_string = pkg_string + ' ' + str(pkg)

            output_upgrade.update({"packages":pkg_string})

            with open(OUTPUT_JSON, 'w', encoding='utf-8') as f:
                json.dump(output_upgrade, f, ensure_ascii=False, indent=4)

            logging.info("Generate Jsonfile(%s) to complete... ",OUTPUT_JSON)
            logging.info("Update Packages list to config file...")
        except Exception as exc:
            logging.error(exc)

    def _write_local(self):
        try:
            output_upgrade = {}
            output_upgrade.update({"steps_queue":self.steps_queue})
            output_upgrade.update({"_problem_resolver":self._problem_resolver})
            output_upgrade.update({"updating_data":self.updating_data})
            output_upgrade.update({"upgrade_singles":self.upgrade_singles})
            output_upgrade.update({"upgrade_groups":self.upgrade_groups})
            output_upgrade.update({"upgrade_steps":self.upgrade_steps})
            output_upgrade.update({"device_status":self.device_status})
            output_upgrade.update({"shutdown_install":self.shutdown_install})
            output_upgrade.update({"signal_trigger":self.signal_trigger})
            output_upgrade.update({"_available":self._available})

            #6 产生JSON文件
            with open(self.UPDATE_AVAILABLE_DATA, 'w', encoding='utf-8') as f:
                json.dump(output_upgrade, f, ensure_ascii=False, indent=4)
            logging.info("Write Upgrade data to local json(%s) in shutdown model...",self.UPDATE_AVAILABLE_DATA)
        except Exception as exc:
            logging.error(exc)

    def read_local(self):
        #读取组JSON文件
        try:
            with open(self.UPDATE_AVAILABLE_DATA,'r') as f:
                data = json.load(f)
                self.steps_queue = data["steps_queue"]
                self._problem_resolver = data["_problem_resolver"]
                self.updating_data = data["updating_data"]
                self.upgrade_singles = data["upgrade_singles"]
                self.upgrade_groups = data["upgrade_groups"]
                self.upgrade_steps = data["upgrade_steps"]
                self.device_status = data["device_status"]
                self.shutdown_install = data["shutdown_install"]
                self.signal_trigger = data["signal_trigger"]
                self._available = data["_available"]
            logging.info("Finished: reading upgrade data from the local file...")
        except Exception as exc:
            logging.error(exc)

    def _get_available(self):
        return self._available

    def _set_available(self, state):
        self._available = state
        self._write_local()

    available = property(_get_available, _set_available)

    # def get_pre_script(self,content):
    #     script_list = []

    #     if content == self.resolver_name:
    #         for groups in self._push_groups:
    #             content = groups
    #             break

    #     script = self._pkgs_data.get(content,{}).get("script",None)
    #     if script == None:
    #         return script_list
        
    #     for scr in ["pre_update_status_check","pre_update_script"]:
    #         run_path = script.get(scr,[])

    #         if isinstance(run_path,list) == True:
    #             for sc in run_path:
    #                 if sc != '':
    #                     script_list.append(sc)
    #         elif isinstance(run_path,str) == True:
    #             if run_path != '':
    #                 script_list.append(run_path)
    #         else:
    #             pass

    #     return script_list

    # def get_post_script(self,content):
    #     script_list = []
    #     if content == self.resolver_name:
    #         for groups in self._push_groups:
    #             content = groups
    #             break

    #     script = self._pkgs_data.get(content,{}).get("script",None)
    #     if script == None:
    #         return script_list
        
    #     for scr in ["post_update_status_check","post_update_script"]:
    #         run_path = script.get(scr,[])

    #         if isinstance(run_path,list) == True:
    #             for sc in run_path:
    #                 if sc != '':
    #                     script_list.append(sc)
    #         elif isinstance(run_path,str) == True:
    #             if run_path != '':
    #                 script_list.append(run_path)
    #         else:
    #             pass

    #     return script_list