#!/usr/bin/env python

import aptdaemon.client as client
import aptdaemon.errors as errors
import aptdaemon.enums as enums
from SystemUpdater.Core.errors import *
from SystemUpdater.Core.enums import *
from aptdaemon.enums import (
                            EXIT_SUCCESS,
                            EXIT_FAILED,
                            EXIT_CANCELLED,
                            STATUS_FINISHED,
                            ERROR_PACKAGE_DOWNLOAD_FAILED,
                            get_status_string_from_enum
                            )

from aptdaemon.enums import get_error_description_from_enum as get_other_error_description_from_enum
from aptdaemon.enums import get_error_string_from_enum as get_other_error_string_from_enum

from defer import inline_callbacks
from SystemUpdater.backend import InstallBackend
import logging
from gettext import gettext as _
from SystemUpdater.Core.utils import humanize_size
import dbus,time
from gi.repository import GLib
import traceback
from importlib import reload

UPDATER_IDLE_CHECK_INTERVAL = 5
INSTALL_IDLE_TIMEOUT = 700
UPDATE_IDLE_TIMEOUT = 5 * 60

class InstallBackendAptdaemon(InstallBackend):
    """Makes use of aptdaemon to refresh the cache and to install updates."""

    def __init__(self, window_main, action,mode = InstallBackend.MODE_DEFAULT_STATUS,data=None):
        InstallBackend.__init__(self, window_main, action,mode,data)
        self.window_main = window_main
        if self.window_main.aptd_lang_switch == True:
            self.window_main.aptd_lang_switch = False
            reload(client)
            reload(errors)
            reload(enums)

        self.client = client.AptClient()

        if action == self.ACTION_UPDATE:
            self.update_timestamp = 0
            GLib.timeout_add_seconds(UPDATER_IDLE_CHECK_INTERVAL,
                self._check_update_inactivity)
            
            self.simulation_progress = 0
            self.tid = GLib.timeout_add_seconds(1,self._check_simulation_progress)
        
        elif action == self.ACTION_INSTALL or action == self.ACTION_INSTALL_SHUTDOWN:
            self.install_timestamp = INSTALL_IDLE_TIMEOUT
            self.check_progress = 0
            GLib.timeout_add_seconds(UPDATER_IDLE_CHECK_INTERVAL,
                self._check_install_inactivity)
        
    def _check_simulation_progress(self):
        self.simulation_progress += 20

        if self.aptd_base.progress >= 90 or self.simulation_progress > 80:
            return False
        else:
            self.aptd_base.progress = self.simulation_progress
            self._dist_status_changed(self.action)
            return True

    def _check_install_inactivity(self):
        """Shutdown the daemon if it has been inactive for time specified
        in INSTALL_IDLE_TIMEOUT.
        """
        logging.info("Checking for inactivity in Installing Time:%d...",self.install_timestamp)

        if self.window_main.now_working != self.ACTION_INSTALL and self.window_main.now_working != self.ACTION_INSTALL_SHUTDOWN:
            logging.info("Installing to exit and timeout check quit...")
            return False

        if self.aptd_base.progress !=  self.check_progress:
            self.check_progress = self.aptd_base.progress
            self.install_timestamp = INSTALL_IDLE_TIMEOUT

        if (self.install_timestamp <= 0 and self.aptd_base.on_installing == True):
            logging.error("Quitting due to inactivity(%s)",self.aptd_base.details)
            self._action_done(self.action,
                            is_cancelled=False, success=False,
                            error_string=_("Could not install the upgrades"), 
                            error_desc=_("Installtion timeout to exit Due to inactivity") + self.aptd_base.details) 

            # self.dbus_send.Quit(None)
            return False
        else:
            self.install_timestamp = self.install_timestamp - 1

        return True

    def _check_update_inactivity(self):
        logging.info("Checking for inactivity in Updating...")
        if (self.aptd_base.cancelable == False and self.aptd_base.progress >= 90) or self.window_main.now_working != self.ACTION_UPDATE:
            logging.info("Updating to exit and timeout check quit...")
            return False

        timestamp = self.update_timestamp
        if self.aptd_base.cancelable == True and timestamp == 0:
            self.update_timestamp = time.time()
            return True
        
        if self.update_timestamp != 0 and  time.time() - self.update_timestamp > UPDATE_IDLE_TIMEOUT \
            and self.aptd_base.cancelable == True:
            logging.error("Quitting due to inactivity")
            self.dbus_send.transaction.cancel()
            return False
        
        return True

    @inline_callbacks
    def update(self):
        """刷新包cache"""
        try:
            trans = yield self.client.update_cache(defer=True,sources_list="test")
            self.dbus_send.transaction = trans
            # 注册回调函数 接收更新的状态
            yield self._show_transaction(trans, self.ACTION_UPDATE,
                                        _("Checking for updates…"), False)
        except errors.NotAuthorizedError:
            self._action_done(self.ACTION_UPDATE,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except Exception:
            self._action_done(self.ACTION_UPDATE,
                            is_cancelled=False, success=False,
                            error_string='', error_desc='')
            raise
    
    # 使用系统源 进行安装
    @inline_callbacks
    def commit(self,action,pkgs_install=[], pkgs_upgrade=[], pkgs_remove=[],pkgs_downgrade = []):
        """Commit a list of package adds and removes"""
        try:
            reinstall = purge = []
            trans = yield self.client.commit_packages(
                pkgs_install, reinstall, pkgs_remove, purge = purge, upgrade = pkgs_upgrade,
                downgrade = pkgs_downgrade,defer=True)
            self.dbus_send.transaction = trans
            
            yield self._show_transaction(trans, action,
                                        _("Installing updates…"), True)
        except errors.NotAuthorizedError:
            self._action_done(action,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except errors.TransactionFailed as e:
            logging.error(str(e))
        except dbus.DBusException as e:
            if e.get_dbus_name() != "org.freedesktop.DBus.Error.NoReply":
                raise
            self._action_done(action,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except Exception:
            self._action_done(action,
                            is_cancelled=False, success=False,
                            error_string='', error_desc='')
            raise
    # 使用更新源 此接口进度直接从50开始 下载过程很短 目前仅优先升级使用
    @inline_callbacks
    def commit_next(self,action,pkgs_install=[], pkgs_upgrade=[], pkgs_remove=[],pkgs_downgrade = []):
        """Commit a list of package adds and removes"""
        try:
            reinstall = purge = []
            trans = yield self.client.commit_only(
                pkgs_install, reinstall, pkgs_remove, purge = purge, upgrade = pkgs_upgrade,
                downgrade = pkgs_downgrade,download = self.ACTION_INSTALL_SHUTDOWN,defer=True)
            self.dbus_send.transaction = trans
            
            yield self._show_transaction(trans, action,
                                        _("Installing updates…"), True)
        except errors.NotAuthorizedError:
            self._action_done(action,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except errors.TransactionFailed as e:
            logging.error(str(e))
        except dbus.DBusException as e:
            if e.get_dbus_name() != "org.freedesktop.DBus.Error.NoReply":
                raise
            self._action_done(action,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except Exception:
            self._action_done(action,
                            is_cancelled=False, success=False,
                            error_string='', error_desc='')
            raise

    # 使用系统更新
    @inline_callbacks
    def commit_download(self,pkgs_install, pkgs_upgrade, pkgs_remove):
        """Commit a list of package adds and removes"""
        try:
            reinstall = purge = downgrade = []
            trans = yield self.client.commit_only(
                pkgs_install, reinstall, pkgs_remove, purge = purge, upgrade = pkgs_upgrade,\
                                downgrade = downgrade,download = self.ACTION_DOWNLOADONLY, defer=True)
            self.dbus_send.transaction = trans
            
            yield self._show_transaction_download(trans, self.ACTION_DOWNLOADONLY,
                                        _("Installing updates…"))
        except errors.NotAuthorizedError:
            self._action_done(self.ACTION_DOWNLOADONLY,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except errors.TransactionFailed as e:
            logging.error(str(e))
        except dbus.DBusException as e:
            if e.get_dbus_name() != "org.freedesktop.DBus.Error.NoReply":
                raise
            self._action_done(self.ACTION_DOWNLOADONLY,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except Exception:
            self._action_done(self.ACTION_DOWNLOADONLY,
                            is_cancelled=False, success=False,
                            error_string='', error_desc='')
            raise
    
    # 使用系统更新
    @inline_callbacks
    def commit_install(self,pkgs_install, pkgs_upgrade, pkgs_remove,final_stage=False):
        """Commit a list of package adds and removes"""
        try:
            reinstall = purge = downgrade =  []
            trans = yield self.client.commit_only(
                pkgs_install, reinstall, pkgs_remove, purge = purge, upgrade = pkgs_upgrade,
                downgrade = downgrade,download = self.ACTION_INSTALL_SHUTDOWN, defer=True)
            self.dbus_send.transaction = trans
            
            yield self._show_transaction_steps(trans, self.action,
                                        _("Installing updates…"), final_stage)
                
        except errors.NotAuthorizedError:
            self._action_done(self.action,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except errors.TransactionFailed as e:
            logging.error(str(e))
        except dbus.DBusException as e:
            if e.get_dbus_name() != "org.freedesktop.DBus.Error.NoReply":
                raise
            self._action_done(self.ACTION_INSTALL_SHUTDOWN,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except Exception:
            self._action_done(self.ACTION_INSTALL_SHUTDOWN,
                            is_cancelled=False, success=False,
                            error_string='', error_desc='')
            raise

    @inline_callbacks
    def install_deb(self,install_path,install_force):
        """安装deb包 """
        try:
            trans = yield self.client.install_file(path = install_path,force = install_force,defer=True)
            # 注册回调函数 接收更新的状态
            yield self._show_transaction(trans, self.ACTION_INSTALL_DEB,
                                        _("Installing deb packages…"), False)
        except errors.NotAuthorizedError:
            self._action_done(self.ACTION_INSTALL_DEB,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except Exception as e:
            self._action_done(self.ACTION_INSTALL_DEB,
                            is_cancelled=False, success=False,
                            error_string=str(e), error_desc='', error_code=e.code)
            # raise

    @inline_callbacks
    def fix_broken(self):
        """安装deb包 """
        try:
            trans = yield self.client.fix_broken_depends(defer=True)
            self.dbus_send.transaction = trans
            # 注册回调函数 接收更新的状态
            yield self._show_transaction(trans, self.ACTION_FIX_BROKEN,
                                        _("Installing deb packages…"), False)
        except errors.NotAuthorizedError:
            self._action_done(self.ACTION_FIX_BROKEN,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except Exception:
            self._action_done(self.ACTION_FIX_BROKEN,
                            is_cancelled=False, success=False,
                            error_string='', error_desc='')
            raise
    
    @inline_callbacks
    def fix_incomplete(self):
        """修复未完成的安装 """
        try:
            trans = yield self.client.fix_incomplete_install(defer=True)
            self.dbus_send.transaction = trans
            # 注册回调函数 接收更新的状态
            yield self._show_transaction(trans, self.ACTION_FIX_INCOMPLETE,
                                        _("fix incomplete install"), False)
        except errors.NotAuthorizedError:
            self._action_done(self.ACTION_FIX_INCOMPLETE,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except Exception:
            self._action_done(self.ACTION_FIX_INCOMPLETE,
                            is_cancelled=False, success=False,
                            error_string='', error_desc='')
            raise

    @inline_callbacks
    def clean(self):
        """清空所有下载的文件 """
        try:
            trans = yield self.client.clean(defer=True)
            self.dbus_send.transaction = trans
            # 注册回调函数 接收更新的状态
            yield self._show_transaction(trans, self.ACTION_CLEAN,
                                        _("Remove all downloaded files."), False)
        except errors.NotAuthorizedError:
            self._action_done(self.ACTION_CLEAN,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except Exception:
            self._action_done(self.ACTION_CLEAN,
                            is_cancelled=False, success=False,
                            error_string='', error_desc='')
            raise

    @inline_callbacks
    def purge_packages(self,pkgs_purge):
        """卸载deb包 """
        try:
            trans = yield self.client.commit_packages([],[],[],pkgs_purge,[],[],defer=True)
            self.dbus_send.transaction = trans
            # 注册回调函数 接收更新的状态
            yield self._show_transaction(trans, self.ACTION_REMOVE_PACKAGES,
                                        _("Installing deb packages…"), False)
        except errors.NotAuthorizedError:
            self._action_done(self.ACTION_REMOVE_PACKAGES,
                            authorized=False, success=False,
                            error_string='', error_desc='')
        except Exception as e:
            logging.error(str(e))

    #进度回调
    def _on_progress_changed(self, trans,progress,action):
        #不要101这种未知状态
        if progress == 101:
            return
        self.aptd_base.progress = progress
        self._dist_status_changed(action)

    #同步状态回调
    def _on_status_changed(self, trans, status,action):
        if action == self.ACTION_UPDATE and status == STATUS_FINISHED:
            return

        #转化词条
        self.aptd_base.status = get_status_string_from_enum(status)
        if self.aptd_base.status == None:
            return

        self._dist_status_changed(action)

    def _on_details_changed(self, trans, details,action):
        self.aptd_base.details = details
        self._dist_status_changed(action)

    # eta 剩余时间不正确，取消掉
    def _on_progress_download_changed(self,trans,current_items, total_items, currenty_bytes, total_bytes, current_cps, eta,action):
        if action == self.ACTION_INSTALL or action == self.ACTION_DOWNLOADONLY or action == self.ACTION_BACKGROUND_UPGRADE:
            if action == self.ACTION_INSTALL or action == self.ACTION_DOWNLOADONLY:
                self.aptd_base.current_items = current_items
                self.aptd_base.total_items = total_items
            self.dbus_send.UpdateDownloadInfo(self.upgrading_content,\
                            current_items,  total_items, \
                            currenty_bytes, total_bytes, \
                            current_cps)
        else:
            if action == self.ACTION_UPDATE or action == self.ACTION_REMOVE_PACKAGES:
                return
            logging.info("Other Action:current_items = %d, total_items = %d, currenty_bytes = %s, total_bytes = %s, current_cps = %s/s",\
            current_items, total_items, \
            humanize_size(currenty_bytes), humanize_size(total_bytes),\
            humanize_size(current_cps))

    def _on_cancellable_changed(self, trans, Cancelable,action):
        #下面的这些 不发送取消信号
        if action == self.ACTION_REMOVE_PACKAGES:
            return

        if action != self.ACTION_UPDATE:
            logging.info("\033[1;32m" + "Emitting" + "\033[0m" +" Cancelable: %r",Cancelable) 
        self.dbus_send.Cancelable(Cancelable)
        self.aptd_base.cancelable = Cancelable

    def _on_config_file_conflict(self, transaction, old, new):
        logging.info("Config file conflict oldconf = %s , newconf = %s...",str(old),str(new))
        logging.info("Default To Replace Old Configfile...")

        transaction.resolve_config_file_conflict(old, "keep")

    #增加记录当产生错误的时候 详细信息 
    def _on_error_changed(self, trans,error_code, error_details):
        self.aptd_base.error_details = str(error_details)
        logging.error(str(error_details))

    #分发进度状态和细节信息
    def _dist_status_changed(self,action):
        progress = self.aptd_base.progress
        status = self.aptd_base.status
        details = self.aptd_base.details

        action_mode = self.action_mode

        if action == self.ACTION_UPDATE:
            if progress == 11:
                progress = 15

            if progress != 100:
                if action_mode == self.MODE_UPDATE_ALL:
                    self.dbus_send.UpdateDetectStatusChanged(progress,status)
                elif action_mode == self.MODE_UPDATE_CACHE:
                    self.dbus_send.UpdatCacheStatusChanged(progress,status)
                
        elif action == self.ACTION_INSTALL:
            #在下载阶段发送取消信号
            if self.aptd_base.on_installing == False:
                self.dbus_send.Cancelable(self.aptd_base.cancelable)
            self.dbus_send.UpdateDloadAndInstStaChanged(self.upgrading_content,progress,status,details)
        elif action == self.ACTION_DOWNLOADONLY:
            self.dbus_send.Cancelable(self.aptd_base.cancelable)
            self.dbus_send.UpdateDloadAndInstStaChanged(self.upgrading_content,progress,status,details)
        elif action == self.ACTION_INSTALL_SHUTDOWN:
            # 写入进度 到plymouth
            self._progress_to_plymouth(progress)
            mesa = _("The system is installing updates. Please do not turn off your computer. ")
            self._message_to_plymouth(mesa)
            self.dbus_send.UpdateDloadAndInstStaChanged(self.upgrading_content,progress,status,details)
        elif action == self.ACTION_FIX_BROKEN:
            self.dbus_send.FixBrokenStatusChanged(False,True,progress,status,'','')
        elif action == self.ACTION_REMOVE_PACKAGES:
            self.dbus_send.PurgePkgStatusChanged(progress,status,details)
        elif action == self.ACTION_INSTALL_DEB or action == self.ACTION_BACKGROUND_UPGRADE:
            self.dbus_send.InstalldebStatusChanged(progress,status,details)
        else:
            logging.info("Other Action:progress = %d , status = %s ,details = %s",progress,status,details)

    @inline_callbacks
    def _show_transaction_download(self, trans, action,upgrading_content):
        trans.connect("finished", self._on_finished, action)
        #升级和更新的状态信息和进度
        trans.connect("status-changed", self._on_status_changed,action)
        trans.connect("progress-changed", self._on_progress_changed,action)
        #取消升级
        trans.connect("cancellable-changed", self._on_cancellable_changed,action)
        #下载的进度信息
        trans.connect("progress-details-changed", self._on_progress_download_changed,action)
        trans.connect("status-details-changed", self._on_details_changed,action)
        trans.connect("error", self._on_error_changed)

        trans.connect("config-file-conflict", self._on_config_file_conflict)

        yield trans.run()

    @inline_callbacks
    def _show_transaction_steps(self, trans, action, header, final_stage):
        if final_stage == True:
            trans.connect("finished", self._on_finished, action)
        else:
            trans.connect("finished", self._on_finished_steps, action)
        #升级和更新的状态信息和进度
        trans.connect("status-changed", self._on_status_changed,action)
        trans.connect("progress-changed", self._on_progress_changed,action)
        #取消升级
        trans.connect("cancellable-changed", self._on_cancellable_changed,action)
        #下载的进度信息
        trans.connect("progress-details-changed", self._on_progress_download_changed,action)
        trans.connect("status-details-changed", self._on_details_changed,action)
        trans.connect("error", self._on_error_changed)

        trans.connect("config-file-conflict", self._on_config_file_conflict)

        yield trans.run()

    @inline_callbacks
    def _show_transaction(self, trans, action, header, show_details):
        if action == self.ACTION_INSTALL_DEB:
            trans.connect("finished", self._on_finished_special, action)
        else:
            trans.connect("finished", self._on_finished, action)
        trans.connect("status-changed", self._on_status_changed,action)
        trans.connect("progress-changed", self._on_progress_changed,action)
        #取消升级
        trans.connect("cancellable-changed", self._on_cancellable_changed,action)
        #下载的进度信息
        trans.connect("progress-details-changed", self._on_progress_download_changed,action)
        trans.connect("status-details-changed", self._on_details_changed,action)
        trans.connect("error", self._on_error_changed)

        trans.connect("config-file-conflict", self._on_config_file_conflict)

        yield trans.run()

    def _on_finished_steps(self, trans, status, action):
        try:
            if status == EXIT_SUCCESS:
                self._start_install_queue()
            else:
                self._on_finished(trans, status, action)
        except Exception as e:
            logging.error(e)
            tbk = traceback.format_exc()
            logging.error(str(tbk))

    def _on_finished(self, trans, status, action):
        try:
            error_string = ''
            error_desc = ''
            error_code = ''
            is_success = (status == EXIT_SUCCESS)
            
            if status == EXIT_FAILED:
                error_code = trans.error.code
                error_string = get_other_error_string_from_enum(error_code)
                error_desc = get_other_error_description_from_enum(error_code)
            elif status == EXIT_CANCELLED:
                error_code = EXIT_CANCELLED
                error_string = get_error_string_from_enum(error_code)
                error_desc = get_error_description_from_enum(error_code)
            elif status == EXIT_SUCCESS and action == self.ACTION_REMOVE_PACKAGES:
                error_string = _("Uninstallation completed")

            self._action_done(action,
                            is_cancelled=(status == EXIT_CANCELLED), success=is_success,
                            error_string=error_string, error_desc=error_desc,error_code=error_code)
        except Exception as e:
            logging.error(e)
            tbk = traceback.format_exc()
            logging.error(str(tbk))
    
    def _on_finished_special(self, trans, status, action):
        try:
            error_string = ''
            error_desc = ''
            error_code = ''

            if status == EXIT_FAILED and action == self.ACTION_INSTALL_DEB:
                error_code = trans.error.code
                error_string = get_error_string_from_enum(trans.error.code)
                error_desc = get_error_description_from_enum(trans.error.code)
                if str(error_code) == 'error-dep-resolution-failed':
                    status,detailed_information,err_info = self.window_main.start_deb_resolver(self.database.deb_metadata['deblist'])

            is_success = ((status == EXIT_SUCCESS) or (status == 0))
            self._action_done(action,
                            is_cancelled=(status == EXIT_CANCELLED), success=is_success,
                            error_string=error_string, error_desc=error_desc,error_code=error_code)
        except Exception as e:
            logging.error(e)
            tbk = traceback.format_exc()
            logging.error(str(tbk))
