#!/usr/bin/env python
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-

"""Integration of package managers into SystemUpdater"""

import os
import json
import logging
import subprocess
import traceback
import dbus
import fcntl
import shutil
import locale
import apt_pkg
from apt import Cache
from gettext import gettext as _
from SystemUpdater.Core.errors import *
from SystemUpdater.Core.enums import *
from SystemUpdater.Core.PluginManager import *
from SystemUpdater.Core.DataAcquisition import get_east_8_time
from SystemUpdater.Core.UpdateEssentialItem import UpdateEssentialItem
from aptdaemon.enums import ERROR_PACKAGE_DOWNLOAD_FAILED,ERROR_DAEMON_DIED
from SystemUpdater.Core.utils import is_offline_source

class AptdBaseInformation:
    def __init__(self,
                progress = 0,
                status='',
                details='',
                cancelable = False,
                error_details = '',
                current_items = 1,
                total_items = 1,
                on_installing = False,
                retry_limit = 2
                ):

        self._progress = progress
        self.status = status
        self.details = details
        self.cancelable = cancelable
        self.error_details = error_details
        self.current_items = current_items
        self.total_items = total_items
        self.on_installing = on_installing
        self.retry_limit = retry_limit

        self.start_progress = 0
        self.end_progress = 100

    def _get_progress(self):
        return self._progress

    def _set_progress(self, progress):
        if self.current_items == 0 and \
                self.total_items == 0 and self.on_installing == False:
            return

        if self.start_progress == 0 and self.end_progress == 100:
            if progress > self._progress:
                self._progress = progress
            else:
                return
        else:
            progress = int(self.start_progress + (progress / 100) *
                        (self.end_progress - self.start_progress))
            self._progress = progress

    progress = property(_get_progress, _set_progress,
                    doc="The required disk space of the transaction.")

class InstallBackend(GObject.GObject):

    __gsignals__ = {"finished": (GObject.SIGNAL_RUN_FIRST,
                                GObject.TYPE_NONE,
                                (   GObject.TYPE_BOOLEAN,
                                    GObject.TYPE_STRING
                                )),
                    }

    ACTION_DEFUALT_STATUS = -1
    ACTION_UPDATE = 0
    ACTION_INSTALL = 1
    ACTION_INSTALL_DEB = 2
    ACTION_CHECK_RESOLVER = 3
    ACTION_DOWNLOADONLY = 4
    ACTION_FIX_BROKEN = 5
    ACTION_REMOVE_PACKAGES = 6
    ACTION_FIX_INCOMPLETE = 7
    ACTION_CLEAN = 8
    ACTION_INSTALL_SHUTDOWN = 9
    ACTION_BACKGROUND_UPGRADE = 10
    ACTION_INSTALL_DEB_RESOLVER = 11
    ACTION_SLEF_UPGRADE = 12

    MODE_DEFAULT_STATUS = -1
    
    MODE_INSTALL_PARTIAL = 0
    MODE_INSTALL_ALL = 1
    MODE_INSTALL_SYSTEM = 2

    MODE_UPDATE_CACHE = 0
    MODE_UPDATE_ALL = 1

    MODE_BACKGROUND_UPGRADE = 1
    MODE_OTHER = 2

    def __init__(self, window_main, action,mode = MODE_DEFAULT_STATUS,data = None):
        GObject.GObject.__init__(self)
        self.upgrading_content = []

        self.window_main = window_main
        self.bus = window_main.bus
        self.cache = window_main.cache
        self.action = action
        self.action_mode = mode

        self.window_main.now_working = action

        self.aptd_base = AptdBaseInformation()
        self.collector = window_main.collector

        self.backend_data = data
        self.dbus_send = window_main.dbusController
        self.database = window_main.sqlite3_server
        self.inhibit_shutdown = InhibitShutdownLock()
        
        if action != self.ACTION_UPDATE:
            self.update_essential = UpdateEssentialItem()
            self.scripts = pluginManagerClass()
        
    def start(self,push_content = []):
        try:
            self.upgrading_content = push_content
            if self.action == self.ACTION_SLEF_UPGRADE:
                pkgs_install,pkgs_upgrade,pkgs_remove,pkgs_downgrade = self.get_mark_from_cache(self.cache)

                logging.info("SELF-UPGRADE install:%d , upgrade:%d remove:%d",len(pkgs_install),len(pkgs_upgrade),len(pkgs_remove))
                
                self.update_essential.check_remove_essential(pkgs_remove)
                self.commit_next(self.action,pkgs_install, pkgs_upgrade, pkgs_remove,pkgs_downgrade)
            #下载前的准备
            elif self.action == self.ACTION_DOWNLOADONLY:
                #0、准备数据
                pkgs_install,pkgs_upgrade,pkgs_remove = self.backend_data.get_resolver_upgrade()
                required_download = self.backend_data.get_required_download()
                required_install = self.backend_data.get_required_install()
                self.aptd_base.end_progress = self.backend_data.get_download_progress()

                #1、检查卸载重要软件包
                self.update_essential.check_remove_essential(pkgs_remove)

                if len(pkgs_install) == 0 and len(pkgs_upgrade) == 0:
                    raise UpdateBaseError(ERROR_NOT_UPGRADE_PACKAGES)

                #2、检查磁盘空间
                self._check_free_space(required_download,required_install)
                # for ul in self.upgrading_content:
                #     self.collector.Upgrade_Process_Msg(self.action, {"appname":ul})
                logging.info("INSTALL install:%d , upgrade:%d remove:%d",len(pkgs_install),len(pkgs_upgrade),len(pkgs_remove))
                self.commit_download(pkgs_install, pkgs_upgrade, pkgs_remove)
            elif self.action == self.ACTION_INSTALL or self.action == self.ACTION_INSTALL_SHUTDOWN:
                # 在启动Plymouth界面直接做完检查 有问题直接弹窗提示用户
                
                self.dbus_send.check_device_power()
                
                if self.backend_data.offline_source:
                    if is_offline_source() != True:
                        # 抛出异常ERROR_NO_OFFLINE_SOURCE
                        raise UpdateBaseError(ERROR_NO_OFFLINE_SOURCE)

                if self.action == self.ACTION_INSTALL_SHUTDOWN:
                    self._check_free_space(required_install = self.backend_data.get_required_install())
                    self._start_plymouth_splash()
                    self.check_plymouth_state()

                for scr in self.backend_data.get_pre_script():
                    self.scripts.run_plugin(scr)

                self._get_install_lock(_("Kylin System Updater"),inhibit_sleep = True)

                logging.info("The task Queue(%r) to be upgraded...",self.backend_data.steps_queue)
                logging.info("Final State(%s) of the device after installation is completed...",self.backend_data.device_status)
                pkgs_install,pkgs_upgrade,pkgs_remove = self.backend_data.get_resolver_steps(self.backend_data.resolver_name)

                self._upgrade_content_for_safe(pkgs_install,pkgs_upgrade,pkgs_remove)
                self._start_install_queue()
            elif self.action == self.ACTION_FIX_BROKEN:
                pkgs_install,pkgs_upgrade,pkgs_remove = self._get_pkgs_from_cache(self.cache)

                self.update_essential.check_remove_essential(pkgs_remove)
                self.fix_broken()
            elif self.action == self.ACTION_FIX_INCOMPLETE:
                self.fix_incomplete()
            elif self.action == self.ACTION_CLEAN:
                self.clean()
            elif self.action == self.ACTION_UPDATE:
                self._check_free_space()
                self.update()
        except UpdateBaseError as excep:
            self._action_done(self.action,True,False,excep.header,excep.desc,excep.code)
        except UpdateProgressExit as excep:
            pass
        except Exception as e:
            logging.error(e) 
            tbk = traceback.format_exc()
            logging.error(str(tbk))
        
    def start_alone(self,push_content = [],_is_install = False, caller=''):
        # 安装本地deb包的接口
        self.upgrading_content = push_content
        
        if self.action == self.ACTION_INSTALL_DEB: 
            try:
                self._get_install_lock(caller=caller)
                self.install_deb(install_path = push_content, install_force = _is_install)
            except Exception as e:
                logging.error(str(e))
        elif self.action == self.ACTION_REMOVE_PACKAGES:
            self._get_install_lock(_("Kylin System Updater"))
            self.purge_packages(push_content)
        elif self.action == self.ACTION_INSTALL_DEB_RESOLVER:
            try:
                self.window_main.refresh_cache(False,True)
                pkgs_install,pkgs_upgrade,pkgs_remove = self.window_main.cache.make_problem_resolver(pkgs_install = push_content)
                self._get_install_lock(caller=caller)
                self.commit(self.ACTION_INSTALL,pkgs_install,pkgs_upgrade,pkgs_remove)
            except Exception as e:
                logging.error(str(e))
        # 安装在线包的接口
        elif self.action == self.ACTION_BACKGROUND_UPGRADE:
            try:
                pkgs_installs                = []
                pkgs_upgrade                 = []
                pkgs_remove                  = []
                pkgs_install_with_version    = []
                pkgs_install_without_version = []

                pkgs_install = [ str(pkg) for pkg in push_content]
                logging.info("Install deb package, open cache %d", self.action_mode)
                if self.action_mode == self.MODE_BACKGROUND_UPGRADE:
                    self.window_main.refresh_cache(True,True)
                elif self.action_mode == self.MODE_OTHER:
                    self.window_main.refresh_cache(False,True)
                for pkg_name in pkgs_install:
                    if pkg_name not in self.window_main.cache:
                        if "=" not in pkg_name or ("=" in pkg_name and pkg_name.split("=")[0] not in self.window_main.cache):
                        # 没找到包或格式不正确
                            self.dbus_send.UpdateInstallFinishedNext(False,True,pkgs_install,ERROR_NO_PACKAGE_CACHE)
                            return
                    
                    if "=" in pkg_name:
                        pkgs_install_with_version.append(pkg_name)
                    else:
                        pkgs_install_without_version.append(pkg_name)

                if (len(pkgs_install_without_version)!=0):
                    pkgs_installs,pkgs_upgrade,pkgs_remove = self.window_main.cache.make_problem_resolver(pkgs_install = pkgs_install_without_version)
                    logging.info("background install list: "+",".join(pkgs_installs)+" , *** upgrade list: "+",".join(pkgs_upgrade))
                    if len(pkgs_remove) != 0:
                        logging.info("background remove list: "+",".join(pkgs_remove)+" , exit.")
                        self.dbus_send.UpdateInstallFinishedNext(False,True,pkgs_install_without_version,"kylin-background-upgrade need remove:"+",".join(pkgs_remove))
                        return

                self._get_install_lock(caller=caller)
                self.commit(self.ACTION_BACKGROUND_UPGRADE,pkgs_installs+pkgs_install_with_version,pkgs_upgrade,pkgs_remove)
            except Exception as e:
                self._action_done(self.action,True,False,e.header,e.desc,e.code)

    def update(self):
        """Run a update to refresh the package list"""
        raise NotImplementedError

    def clean(self):
        """Remove all downloaded files.t"""
        raise NotImplementedError

    def commit(self,model,pkgs_install, pkgs_upgrade, pkgs_remove,pkgs_downgrade = []):
        """Commit the cache changes """
        raise NotImplementedError

    def install_deb(self,install_path,install_force):
        """install_deb the cache changes """
        raise NotImplementedError

    def fix_broken(self):
        """fix_broken_depends"""
        raise NotImplementedError

    def fix_incomplete(self):
        """fix_incomplete_install"""
        raise NotImplementedError

    def purge_packages(self,pkgs_purge):
        """purge_packages"""
        raise NotImplementedError

    def _get_pkgs_from_cache(self,cache):
        pkgs_install = []
        pkgs_upgrade = []
        pkgs_remove = []
        for pkg in cache:
            try:
                if pkg.marked_install :
                    pkgs_install.append(pkg.name)
                elif pkg.marked_upgrade:
                    pkgs_upgrade.append(pkg.name)
                elif pkg.marked_delete:
                    pkgs_remove.append(pkg.name)
            except KeyError:
                # pkg missing from fresh_cache can't be modified
                pass
        return pkgs_install,pkgs_upgrade,pkgs_remove

    def get_mark_from_cache(self,cache):
        pkgs_install = []
        pkgs_upgrade = []
        pkgs_remove = []
        pkgs_downgrade = []
        
        for pkg in cache:
            try:
                if pkg.marked_install:
                    pkgname = pkg.name
                    if pkg.is_auto_installed:
                        pkgname += "#auto"
                        pkgs_install.append(pkgname)
                    else:
                        pkgs_install.append(pkgname)
                elif pkg.marked_upgrade:
                        pkgs_upgrade.append(pkg.name)
                elif pkg.marked_delete:
                    pkgs_remove.append(pkg.name)
                elif pkg.marked_downgrade:
                    pkgs_downgrade.append(pkg.name+"="+getattr(pkg.candidate, "version", ''))
            except KeyError:
                pass
        return pkgs_install,pkgs_upgrade,pkgs_remove,pkgs_downgrade

    #检查是否需要重新下载
    def _need_retry_download(self):
        def _check_connectivity_pro():
            try:
                logging.info("Start checking the network...")
                obj = self.bus.get_object("org.freedesktop.NetworkManager","/org/freedesktop/NetworkManager")
                interface = dbus.Interface(obj, "org.freedesktop.NetworkManager")
                retval = interface.CheckConnectivity(timeout=15)
            except Exception as e:
                logging.error("Network status is Exception:" + str(e))
                return False
            if retval == 4:
                logging.info("Network status is Success...")
                return True
            else:
                logging.info("Network status(%d) is Exception...",retval)
                return False
        
        if _check_connectivity_pro() == True and self.aptd_base.retry_limit != 0:
            self.aptd_base.retry_limit = self.aptd_base.retry_limit - 1
            logging.warning("Retry the upgrade and installaton and retry number: %d",self.aptd_base.retry_limit)
            pkgs_install,pkgs_upgrade,pkgs_remove = self.backend_data.get_resolver_upgrade()
            self.commit_download(pkgs_install, pkgs_upgrade, pkgs_remove)
            return True
        else:
            return False

    def _need_retry_install(self):
        if self.aptd_base.retry_limit != 0:
            self.aptd_base.retry_limit = self.aptd_base.retry_limit - 1
            logging.warning("Retry the upgrade and installaton and retry number: %d",self.aptd_base.retry_limit)
            pkgs_install,pkgs_upgrade,pkgs_remove = self.backend_data.get_resolver_steps(self.backend_data.last_task)

            if self.backend_data.check_steps_reboot(self.backend_data.last_task) or self.backend_data.steps_queue == []:
                final_stage = True

            logging.info("INSTALL(%s) install:%d , upgrade:%d remove:%d",self.backend_data.last_task,len(pkgs_install),len(pkgs_upgrade),len(pkgs_remove))
            self.commit_install(pkgs_install, pkgs_upgrade, pkgs_remove,final_stage)
            return True

    def _start_install_queue(self):
        final_stage = False

        now_steps = self.backend_data.steps_queue.pop()
        self.aptd_base.start_progress,self.aptd_base.end_progress = self.backend_data.get_install_progress(now_steps)
        pkgs_install,pkgs_upgrade,pkgs_remove = self.backend_data.get_resolver_steps(now_steps)

        self.dbus_send.NotifyStepsInstalled(True,self.backend_data.get_mid_notify(now_steps))

        if self.backend_data.check_steps_reboot(now_steps) or self.backend_data.steps_queue == []:
            final_stage = True

        self.backend_data.last_task = now_steps

        logging.info("INSTALL(%s) install:%d , upgrade:%d remove:%d",now_steps,len(pkgs_install),len(pkgs_upgrade),len(pkgs_remove))
        self.commit_install(pkgs_install, pkgs_upgrade, pkgs_remove,final_stage)

    def _action_done(self, action, is_cancelled,success, error_string='',error_desc='',error_code=''):    
        self.window_main.now_working = self.ACTION_DEFUALT_STATUS
        if action == self.ACTION_SLEF_UPGRADE:
            self.emit("finished",success,error_code)
        elif action == self.ACTION_DOWNLOADONLY:
            #1、检查是否需要重试下载
            if success == False and error_code == ERROR_PACKAGE_DOWNLOAD_FAILED:
                if self._need_retry_download():
                    return
            self.emit("finished",success,error_code)
            
        elif action == self.ACTION_INSTALL or action == self.ACTION_INSTALL_SHUTDOWN:
            if success == False and error_code in [ERROR_DEVICE_LOW_BATTERY,ERROR_NOT_DISK_SPACE]:
                self.emit("finished",success,error_code)
                return 
            
            # 在离线更新特定机型下 会出现离线源下载失败的问题
            if success == False and self.aptd_base.error_details != "" and \
                                        "(5: Input/output error)" in self.aptd_base.error_details:
                self._need_retry_install()
                
            #1、解除安装状态
            self._release_install_lock()

            #执行安装后脚本
            for scr in self.backend_data.get_post_script():
                self.scripts.run_plugin(scr)
            
            if success and action == self.ACTION_INSTALL:
                self.dbus_send.start_popup_on_installed(True,self.backend_data.get_post_notify())
                self.dbus_send.CheckRebootRequired("self")

            if action == self.ACTION_INSTALL_SHUTDOWN:
                self.plymouth_quit_retain_splash()
                if success == True:
                    if self.backend_data.steps_queue == []:
                        self.disable_next_steps()
                    else:
                        self.enable_next_steps()
                else:
                    self.disable_next_steps()
            
            self.emit("finished",success,error_code)
        elif action == self.ACTION_UPDATE and self.action_mode == self.MODE_UPDATE_ALL:
            self.emit("finished",success,error_code)
            self.database.insert_into_display("check_time",get_east_8_time()[0:-4])
        elif action == self.ACTION_UPDATE and self.action_mode == self.MODE_UPDATE_CACHE:
            self.dbus_send.UpdateCacheFinished(success,[''],error_string,error_desc)
        elif action == self.ACTION_FIX_BROKEN or action == self.ACTION_FIX_INCOMPLETE:
            self.emit("finished",success,error_code)

        elif action == self.ACTION_REMOVE_PACKAGES:
            self._release_install_lock()
            purge_list = self.database.current_purge_pkgs
            if success == False and 'in the prevent uninstallation of blacklists' in self.aptd_base.error_details \
                and error_code == 'error-package-manager-failed':
                error_string = _("Forbid uninstallation of software packages.")

            self.dbus_send.PurgePackagesFinishedWithPkgname(success,error_string,error_desc,",".join(purge_list))
            self.dbus_send.PurgePackagesFinished(success,error_string,error_desc)
            self.dbus_send.PurgePackagesFinishedWithErrCode(success,error_string,error_desc,",".join(purge_list),get_error_num_from_enum(error_code))
            
        elif action == self.ACTION_INSTALL_DEB or action == self.ACTION_INSTALL_DEB_RESOLVER:
            self._release_install_lock()
            if success == False and '\r\n: \r\n\r\n' in self.aptd_base.error_details:
                error_string = _("Package validation failed and installation was rejected.")
                error_desc = ''
            if error_code != None and str(error_code) != 'error-dep-resolution-failed':
                if error_desc != '':
                    error_string = ",".join((error_string,error_desc))
                self.dbus_send.InstalldebFinished(success,error_string,'')
                self.dbus_send.InstalldebFinishedWithErrCode(success,error_string,'',get_error_num_from_enum(error_code))
                UpdateMsg = {}
                if success:
                    status = 'success'
                    UpdateMsg.update({"errorCode":" "})
                else:
                    status = 'failed'
                    UpdateMsg.update({"errorCode":str(error_string+" "+error_desc)})
                #apt发送数据
                if self.window_main.configs_uncover.getWithDefault("SystemStatus", "upload_installer_log", False) == True:
                    
                    UpdateMsg.update({"appname":str(self.window_main.deb_obj.get("debname","None").split("_")[0])})
                    UpdateMsg.update({"source":str(self.window_main.deb_obj.get("source","kylin-system-updater"))})
                    UpdateMsg.update({"status":str(status)})
                    UpdateMsg.update({"new_version":str(self.window_main.deb_obj.get("debname","None").split("_")[1])})
                    UpdateMsg.update({"old_version":str(self.window_main.deb_obj.get("old_version","None"))})
                    self.collector.Upgrade_Process_Msg(action, UpdateMsg.copy())
                    self.window_main.deb_obj = {}
        elif action == self.ACTION_BACKGROUND_UPGRADE:
            self._release_install_lock()
            UpdateMsg = {}

            if self.action_mode == self.MODE_BACKGROUND_UPGRADE:
                if success:
                    status = 'success'
                else:
                    status = 'failed'
                self.collector.Generate_Msg(self.collector.background_list, self.action_mode)
                for bl in self.collector.background_list:
                    pkg = self.collector.cache[bl]
                    UpdateMsg.update({"appname":str(bl)})
                    UpdateMsg.update({"source":"Kylin Background upgrade"})
                    UpdateMsg.update({"status":str(status)})
                    UpdateMsg.update({"errorCode":str(error_string+" "+error_desc)})
                    if self.collector.background_version[bl]:
                        UpdateMsg.update({"old_version":self.collector.background_version[bl]})
                    UpdateMsg.update({"new_version":self.collector.cache[bl].candidate.source_version})
                    self.collector.Upgrade_Process_Msg(self.action, UpdateMsg.copy())
                    UpdateMsg.clear()
                    self.collector.background_list.clear()

            self.dbus_send.UpdateInstallFinished(success,self.upgrading_content,error_string,error_desc)

    def _check_free_space(self,required_download=0,required_install=0):
        from SystemUpdater.Core.DistUpgradeCache import FreeSpaceRequired
        """
        this checks if we have enough free space on /var, /boot and /usr
        with the given cache

        Note: this can not be fully accurate if there are multiple
            mountpoints for /usr, /var, /boot
        """
        logging.info("Check: Enough free space For system update...")

        class FreeSpace(object):
            " helper class that represents the free space on each mounted fs "
            def __init__(self, initialFree):
                self.free = initialFree
                self.need = 0

        def make_fs_id(d):
            """ return 'id' of a directory so that directories on the
                same filesystem get the same id (simply the mount_point)
            """
            for mount_point in mounted:
                if d.startswith(mount_point):
                    return mount_point
            return "/"
        try:
            # this is all a bit complicated
            # 1) check what is mounted (in mounted)
            # 2) create FreeSpace objects for the dirs we are interested in
            #    (mnt_map)
            # 3) use the  mnt_map to check if we have enough free space and
            #    if not tell the user how much is missing
            mounted = []
            mnt_map = {}
            fs_free = {}
            with open("/proc/mounts") as mounts:
                for line in mounts:
                    try:
                        (what, where, fs, options, a, b) = line.split()
                    except ValueError as e:
                        logging.debug("line '%s' in /proc/mounts not understood (%s)" % (line, e))
                        continue
                    if not where in mounted:
                        mounted.append(where)
            # make sure mounted is sorted by longest path
            mounted.sort(key=len, reverse=True)
            archivedir = apt_pkg.config.find_dir("Dir::Cache::archives")
            for d in ["/", "/usr", "/var", "/boot", archivedir, "/home", "/tmp/"]:
                d = os.path.realpath(d)
                fs_id = make_fs_id(d)
                if os.path.exists(d):
                    st = os.statvfs(d)
                    free = st.f_bavail * st.f_frsize
                else:
                    logging.warning("directory '%s' does not exists" % d)
                    free = 0
                if fs_id in mnt_map:
                    # logging.debug("Dir %s mounted on %s" %
                    #                 (d, mnt_map[fs_id]))
                    fs_free[d] = fs_free[mnt_map[fs_id]]
                else:
                    # logging.debug("Free space on %s: %s" %
                    #                 (d, free))
                    mnt_map[fs_id] = d
                    fs_free[d] = FreeSpace(free)
            del mnt_map
            # logging.debug("fs_free contains: '%s'" % fs_free)

            # we check for various sizes:
            # archivedir is where we download the debs
            # /usr is assumed to get *all* of the install space (incorrect,
            #      but as good as we can do currently + safety buffer
            # /     has a small safety buffer as well
            # sum up space requirements
            for (dir, size) in [(archivedir,required_download),
                                ("/usr", required_install),
                                # plus 50M safety buffer in /usr
                                ("/usr", 50*1024*1024),
                                ("/tmp", 5*1024*1024),   # /tmp for dkms LP: #427035
                                ("/", 50*1024*1024),     # small safety buffer /
                                ]:
                # we are ensuring we have more than enough free space not less
                if size < 0:
                    continue
                dir = os.path.realpath(dir)
                # logging.debug("dir '%s' needs '%s' of '%s' (%f)" % (dir, size, fs_free[dir], fs_free[dir].free))
                fs_free[dir].free -= size
                fs_free[dir].need += size

            # check for space required violations
            required_list = {}
            for dir in fs_free:
                if fs_free[dir].free < 0:
                    # ensure unicode here (LP: #1172740)
                    free_at_least = apt_pkg.size_to_str(float(abs(fs_free[dir].free)+1))
                    if isinstance(free_at_least, bytes):
                        free_at_least = free_at_least.decode(
                            locale.getpreferredencoding())
                    free_needed = apt_pkg.size_to_str(fs_free[dir].need)
                    if isinstance(free_needed, bytes):
                        free_needed = free_needed.decode(
                            locale.getpreferredencoding())
                    # make_fs_id ensures we only get stuff on the same
                    # mountpoint, so we report the requirements only once
                    # per mountpoint
                    required_list[make_fs_id(dir)] = FreeSpaceRequired(free_needed, make_fs_id(dir), free_at_least)
        except Exception as e:
            logging.error("free space check failed %s",str(e))
            # raise exception if free space check fails
        if len(required_list) > 0:
            logging.error("Not enough free space: %s" % [str(i) for i in required_list])
            raise UpdateBaseError(ERROR_NOT_DISK_SPACE)
        return True

    def _get_install_lock(self, caller='kylin-system-updater',inhibit_sleep = False):
        next_reboot_num = self.window_main.configs_uncover.getWithDefault("SystemStatus", "check_abnormal_flag", 0) + 1
        self.window_main.configs_uncover.setValue("SystemStatus","check_abnormal_flag",str(next_reboot_num))
        self.inhibit_shutdown.lock(caller=caller,inhibit_sleep = inhibit_sleep)
        self.aptd_base.on_installing = True

    def _release_install_lock(self):
        self.window_main.configs_uncover.setValue("SystemStatus","check_abnormal_flag",str(0))
        self.inhibit_shutdown.unlock()

    def _message_to_plymouth(self,message):
        subprocess.call(["/bin/plymouth", "message", "--text", message])

    def _progress_to_plymouth(self,progress):
        tmp = ("--progress=%d"%progress)
        run_cmd = ["/bin/plymouth","system-update",tmp]
        subprocess.call(run_cmd)


    def _show_systemd_inhibit(self):
        logging.info("Show systemd inhibit...")
        args = ["systemd-inhibit"]
        try:
            p = subprocess.run(args,timeout=2,stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
            logging.info(str(p.stdout))
        except Exception as exc:
            logging.error(exc)

    def check_plymouth_state(self):
        try:
            args = ["/bin/plymouth","--ping"]
            p = subprocess.run(args,timeout=2,stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
            if p.returncode == 0:
                logging.info("Check: Plymouthd state start to success...")
            else:
                logging.info("Check: Plymouthd Failed to Boot and Restart boot up again...")
                self._start_plymouth_splash()
        except Exception as exc:
            logging.error(exc) 

    def _start_plymouth_splash(self):
        try:
            if os.path.exists("/sbin/plymouthd"):
                if not self._check_device_type():
                    args = ["/sbin/plymouthd", "--mode=boot","--attach-to-session"]
                else:
                    os.environ["DISPLAY"] = ":0"
                    args = ["/sbin/plymouthd", "--mode=shutdown","--attach-to-session"]
                logging.info("Running plymouth:%r",args)
                subprocess.run(args,timeout=3)
            if os.path.exists("/bin/plymouth"):
                subprocess.Popen(["/bin/plymouth", "show-splash","--wait"])
                subprocess.call(["/bin/plymouth","system-update","--progress=0"])

            self._progress_to_plymouth(5)
            mesa = _("The system is installing updates. Please do not turn off your computer. ")
            self._message_to_plymouth(mesa)
        except Exception as exc:
            logging.error(exc)

    def plymouth_quit_retain_splash(self):
        if os.path.exists("/bin/plymouth"):
            subprocess.run(["/bin/plymouth","quit","--retain-splash"])

    def _check_device_type(self):
        try:
            with open("/proc/cpuinfo", 'r') as f:
                for line in f.readlines():
                    if "Hardware" in line and ":" in line:
                        cpuinfos = line.split(':')
                        if "PANGU M900" in cpuinfos[1]:
                            logging.info("Get cpuinfo: PANGU M900")
                            return True
        except Exception as exc:
            logging.error(exc)
        return False
            
    def enable_next_steps(self):
        args = ["systemctl","enable","kylin-upgrade-poweroff.service"]
        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        if p.returncode == 0:
            logging.info("start enable kylin-upgrade-poweroff.service to success...")
            return True
        else:
            logging.info("start enable kylin-upgrade-poweroff.service to failed...")
            return False

    def disable_next_steps(self):
        args = ["systemctl","is-enabled","kylin-upgrade-poweroff.service"]
        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        if p.returncode != 0:
            pass
        else:
            logging.info("kylin-upgrade-poweroff.service has been enabled and need to disable...")
            args = ["systemctl","disable","kylin-upgrade-poweroff.service"]
            p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
            if p.returncode == 0:
                logging.info("start disabled kylin-upgrade-poweroff.service to success...")
                return True
            else:
                logging.info("start disabled kylin-upgrade-poweroff.service to failed...")
                return False

    def _upgrade_content_for_safe(self,pkgs_install,pkgs_upgrade,pkgs_remove):
        INSTALL_OUTPUT_JSON = "/var/lib/kylin-system-updater/json/requires_upgrade_meta.json"
        try:
            output_upgrade = {}

            output_upgrade.update({"pkgs_install":pkgs_install})
            output_upgrade.update({"pkgs_upgrade":pkgs_upgrade})
            output_upgrade.update({"pkgs_remove":pkgs_remove})

            with open(INSTALL_OUTPUT_JSON, 'w', encoding='utf-8') as f:
                json.dump(output_upgrade, f, ensure_ascii=False, indent=4)

            logging.info("Generate Jsonfile(%s) to complete... ",INSTALL_OUTPUT_JSON)
            logging.info("Update Packages list to config file...")
        except Exception as exc:
            logging.error(exc)
            
    def emulate_calcul_broken(self,pkgs):
        args = ["apt-get","-c","/etc/kylin-system-updater/apt_private.conf","install","--simulate"]
        args = args + pkgs

        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        # logging.error(str(p.stdout))
        return p.stdout

    def start_backup_restore(self):
        obj = self.bus.get_object('com.kylin.backup','/')
        interface = dbus.Interface(obj, dbus_interface='com.kylin.backup.manager')
        logging.info("Now start to restore system...")
        interface.autoRestoreForSystemUpdate("root",0,timeout=0.5)

        self.connet_restore_progress()

    def connet_restore_progress(self):
        def restore_handle_signal(x,progress):
            logging.info("Backup restore now current progress:%d...",progress)
            self._progress_to_plymouth(progress)

        obj = self.bus.get_object('com.kylin.backup','/')
        obj.connect_to_signal(
            "sendRate",
            restore_handle_signal,
            dbus_interface="com.kylin.backup.manager")

    def emulate_update(self):
        args = ["apt-get", "update"]

        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        logging.warning(str(p.stdout))

# try aptdaemon
if os.path.exists("/usr/sbin/aptd") \
   and "UPDATE_MANAGER_FORCE_BACKEND_SYNAPTIC" not in os.environ:
    # check if the gtkwidgets are installed as well
    try:
        from .InstallBackendAptdaemon import InstallBackendAptdaemon
    except ImportError:
        logging.exception("importing aptdaemon")

def get_backend(*args, **kwargs):
    """Select and return a package manager backend."""
    # try aptdaemon
    if (os.path.exists("/usr/sbin/aptd")
            and "UPDATE_MANAGER_FORCE_BACKEND_SYNAPTIC" not in os.environ):
        # check if the gtkwidgets are installed as well
        try:
            return InstallBackendAptdaemon(*args, **kwargs)
        except NameError:
            logging.exception("using aptdaemon failed")

    # nothing found, raise
    raise Exception("No working backend found, please try installing "
                    "aptdaemon or synaptic")
class InhibitShutdownLock():
    # 禁止关机锁文件路径
    FILELOCK_PATH = "/tmp/lock/"
    SHUTDOWN_BLOCK_FILELOCK = "kylin-update.lock"
    LOGIND_SLEEP_CONF = "/etc/systemd/sleep.conf.d/"
    LOGIND_SLEEP_POSITION = "/etc/systemd/sleep.conf.d/inhibit-sleep.conf"
    SLEEP_CONF_POSITION = "/var/lib/kylin-system-updater/inhibit-sleep.conf"
    
    def __init__(self):
        self.inhibit_lock = None
        self.pidfile = None

    #安装时禁止关机 进行加锁
    def lock(self, caller='kylin-system-updater',inhibit_sleep = False):
        """
        Send a dbus signal to logind to not suspend the system, it will be
        released when the return value drops out of scope
        """
        if inhibit_sleep == True:
            self._lock_sleep()

        if self.inhibit_lock == None:
            try:
                from gi.repository import Gio, GLib
                logging.info("Start acquiring Shutdown lock(block)...")
                connection = Gio.bus_get_sync(Gio.BusType.SYSTEM)
                
                var, fdlist = connection.call_with_unix_fd_list_sync(
                    'org.freedesktop.login1', '/org/freedesktop/login1',
                    'org.freedesktop.login1.Manager', 'Inhibit',
                    GLib.Variant('(ssss)',
                                ('shutdown:sleep',
                                caller, _("Installing or removing software packages."),
                                'block')),
                    None, 0, -1, None, None)
                self.inhibit_lock = Gio.UnixInputStream(fd=fdlist.steal_fds()[var[0]])
                self.LockedPreventShutdown()
                
                logging.info("Shutdown Has been locked...")
                return True
            except Exception as e:
                logging.error(e)
                return False
        else:
            logging.info("No need to acquired lock and Shutdown Has been locked...")
            return False

    #解锁禁止关机 
    def unlock(self):
        try:
            self.unLockedEnableShutdown()
            self._unlock_sleep()
            if self.inhibit_lock != None:
                self.inhibit_lock.close()
                self.inhibit_lock = None
                logging.info("Shutdown Has been unlocked...")
            else:
                logging.info("Not locked and Quitting ...")
        except Exception as e:
            self.inhibit_lock = None
            logging.error("unlock failed." + str(e))

    def _lock_sleep(self):
        try:
            if not os.path.exists(self.LOGIND_SLEEP_CONF):
                logging.info("File(%s) is not exists and will be create",self.LOGIND_SLEEP_CONF)
                os.makedirs(self.LOGIND_SLEEP_CONF)
            shutil.copy(self.SLEEP_CONF_POSITION, self.LOGIND_SLEEP_CONF)
            logging.info("File(%s) will be create for forcely inhibit sleep...",self.LOGIND_SLEEP_CONF)
        except Exception as e:
            logging.error(e)
            
    def _unlock_sleep(self):
        try:
            if os.path.exists(self.LOGIND_SLEEP_POSITION):
                os.remove(self.LOGIND_SLEEP_POSITION)
                # logging.info("File(%s) is exists and will be delete and unlock sleep",self.LOGIND_SLEEP_POSITION)
        except Exception as e:
            logging.error(e)
    @staticmethod
    def _check_unlock_sleep():
        try:
            LOGIND_SLEEP_POSITION = "/etc/systemd/sleep.conf.d/inhibit-sleep.conf"
            if os.path.exists(LOGIND_SLEEP_POSITION):
                os.remove(LOGIND_SLEEP_POSITION)
                logging.warning("File(%s) is exists and will be delete and unlock sleep",LOGIND_SLEEP_POSITION)
        except Exception as e:
            logging.error(e)

    #安装时禁止关机 进行加锁
    def LockedPreventShutdown(self):
        #不为空是表示以及被锁
        if self.pidfile != None:
            logging.error("self.pidfile file disc not is None,Has been locked...")
            return False

        if not os.path.exists(self.FILELOCK_PATH):
            #不存在创建
            logging.info("File(%s) is not exists and will be create",self.FILELOCK_PATH)
            os.makedirs(self.FILELOCK_PATH)
        else:
            #当目录存在时进行删除 不删除进行创建文件的话会报错 
            # file cannot be locked.[Errno 11] Resource temporarily unavailable
            # 资源被占用报错
            shutil.rmtree(self.FILELOCK_PATH)
            logging.info("File(%s) is exists and will be delete and create",self.FILELOCK_PATH)
            os.makedirs(self.FILELOCK_PATH)        

        try:
            self.pidfile = open(os.path.join(self.FILELOCK_PATH, self.SHUTDOWN_BLOCK_FILELOCK), "w+")
            fcntl.flock(self.pidfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
            return True
        except Exception as e:
            logging.error("file cannot be locked." + str(e))
            self.pidfile = None
            return False

    #解锁禁止关机 
    def unLockedEnableShutdown(self):
        #未加锁退出
        if not self.pidfile:
            logging.info("Not locked and Quitting ...")
            return False
        try:
            fcntl.flock(self.pidfile, fcntl.LOCK_UN)
            self.pidfile.close()
            self.pidfile = None

            # Fix 修复权限问题 当普通用户无法使用 所以直接删除目录
            if os.path.exists(self.FILELOCK_PATH):
                shutil.rmtree(self.FILELOCK_PATH)
                logging.info('Emptying the lockPath(%s) is complete...',self.FILELOCK_PATH)
            else:
                logging.info("Emptying the lockPath(%s) is Failed...",self.FILELOCK_PATH)

            return True
        except Exception as e:
            logging.error("unlock failed." + str(e))
            self.pidfile = None
            return False   